import beignet
import torch


# def test_chebyshev_polynomial_roots():
#     torch.testing.assert_close(
#         beignet.chebyshev_polynomial_roots(
#             torch.tensor([1.0]),
#         ),
#         torch.tensor([]),
#     )
#
#     torch.testing.assert_close(
#         beignet.chebyshev_polynomial_roots(
#             torch.tensor([1.0, 2.0]),
#         ),
#         torch.tensor([-0.5]),
#     )
#
#     for i in range(2, 5):
#         torch.testing.assert_close(
#             beignet.trim_chebyshev_polynomial_coefficients(
#                 beignet.chebyshev_polynomial_roots(
#                     beignet.chebyshev_polynomial_from_roots(
#                         torch.linspace(-1, 1, i),
#                     )
#                 ),
#                 tol=0.000001,
#             ),
#             beignet.trim_chebyshev_polynomial_coefficients(
#                 torch.linspace(-1, 1, i),
#                 tol=0.000001,
#             ),
#         )



import re

import torch
from torch import Tensor


def euler_angle_to_rotation_vector(
    input: Tensor,
    axes: str,
    degrees: bool = False,
) -> Tensor:
    r"""
    Convert Euler angles to rotation vectors.

    Parameters
    ----------
    input : Tensor, shape=(..., 3)
        Euler angles.

    axes : str
        Axes. One to three characters belonging to the set :math:`\{X, Y, Z\}`
        for intrinsic rotations, or :math:`\{x, y, z\}` for extrinsic
        rotations. Extrinsic and intrinsic rotations cannot be mixed.

    degrees : bool, optional
        If `True`, Euler angles are assumed to be in degrees and returned
        rotation vector magnitudes are in degrees. Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 3)
        Rotation vectors.
    """
    num_axes = len(axes)

    if num_axes < 1 or num_axes > 3:
        raise ValueError

    intrinsic = re.match(r"^[XYZ]{1,3}$", axes) is not None
    extrinsic = re.match(r"^[xyz]{1,3}$", axes) is not None

    if not (intrinsic or extrinsic):
        raise ValueError

    if any(axes[i] == axes[i + 1] for i in range(num_axes - 1)):
        raise ValueError

    if degrees:
        input = torch.deg2rad(input)

    if len(axes.lower()) == 1:
        match input.ndim:
            case 0:
                input = torch.reshape(input, [1, 1])
            case 1:
                input = input[:, None]
            case 2 if input.shape[-1] != 1:
                raise ValueError
            case _:
                raise ValueError

    else:
        if input.ndim not in [1, 2] or input.shape[-1] != len(axes.lower()):
            raise ValueError

        if input.ndim == 1:
            input = input[None, :]

    if input.ndim != 2 or input.shape[-1] != len(axes.lower()):
        raise ValueError

    x = torch.zeros(
        [input.shape[0], 4],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    match axes.lower()[0]:
        case "x":
            m = 0
        case "y":
            m = 1
        case "z":
            m = 2
        case _:
            raise ValueError

    for j in range(input[:, 0].shape[0]):
        x[j, 3] = torch.cos(input[:, 0][j] / 2)
        x[j, m] = torch.sin(input[:, 0][j] / 2)

    for j in range(1, len(axes)):
        y = torch.zeros(
            [input.shape[0], 4],
            dtype=input.dtype,
            layout=input.layout,
            device=input.device,
        )

        z = torch.empty(
            [max(y.shape[0], x.shape[0]), 4],
            dtype=input.dtype,
            layout=input.layout,
            device=input.device,
        )

        match axes.lower()[j]:
            case "x":
                m = 0
            case "y":
                m = 1
            case "z":
                m = 2
            case _:
                raise ValueError

        for k in range(input[:, j].shape[0]):
            y[k, 3] = torch.cos(input[:, j][k] / 2)
            y[k, m] = torch.sin(input[:, j][k] / 2)

        if intrinsic:
            if x.shape[0] == 1:
                for k in range(max(x.shape[0], y.shape[0])):
                    q = y[k, 1]
                    r = y[k, 2]
                    s = y[k, 3]
                    p = y[k, 0]

                    t = x[0, 0]
                    u = x[0, 1]
                    v = x[0, 2]
                    w = x[0, 3]

                    z[k, 0] = w * p + s * t + u * r - v * q
                    z[k, 1] = w * q + s * u + v * p - t * r
                    z[k, 2] = w * r + s * v + t * q - u * p
                    z[k, 3] = w * s - t * p - u * q - v * r
            elif y.shape[0] == 1:
                for k in range(max(x.shape[0], y.shape[0])):
                    p = y[0, 0]
                    q = y[0, 1]
                    r = y[0, 2]
                    s = y[0, 3]

                    t = x[k, 0]
                    u = x[k, 1]
                    v = x[k, 2]
                    w = x[k, 3]

                    z[k, 0] = w * p + s * t + u * r - v * q
                    z[k, 1] = w * q + s * u + v * p - t * r
                    z[k, 2] = w * r + s * v + t * q - u * p
                    z[k, 3] = w * s - t * p - u * q - v * r
            else:
                for k in range(max(x.shape[0], y.shape[0])):
                    p = y[k, 0]
                    q = y[k, 1]
                    r = y[k, 2]
                    s = y[k, 3]

                    t = x[k, 0]
                    u = x[k, 1]
                    v = x[k, 2]
                    w = x[k, 3]

                    z[k, 0] = w * p + s * t + u * r - v * q
                    z[k, 1] = w * q + s * u + v * p - t * r
                    z[k, 2] = w * r + s * v + t * q - u * p
                    z[k, 3] = w * s - t * p - u * q - v * r

            x = z
        else:
            if y.shape[0] == 1:
                for k in range(max(y.shape[0], x.shape[0])):
                    p = y[0, 0]
                    q = y[0, 1]
                    r = y[0, 2]
                    s = y[0, 3]

                    t = x[k, 0]
                    u = x[k, 1]
                    v = x[k, 2]
                    w = x[k, 3]

                    z[k, 0] = s * t + w * p + q * v - r * u
                    z[k, 1] = s * u + w * q + r * t - p * v
                    z[k, 2] = s * v + w * r + p * u - q * t
                    z[k, 3] = s * w - p * t - q * u - r * v
            elif x.shape[0] == 1:
                for k in range(max(y.shape[0], x.shape[0])):
                    t = x[0, 0]
                    u = x[0, 1]
                    v = x[0, 2]
                    w = x[0, 3]

                    p = y[k, 0]
                    q = y[k, 1]
                    r = y[k, 2]
                    s = y[k, 3]

                    z[k, 0] = s * t + w * p + q * v - r * u
                    z[k, 1] = s * u + w * q + r * t - p * v
                    z[k, 2] = s * v + w * r + p * u - q * t
                    z[k, 3] = s * w - p * t - q * u - r * v
            else:
                for k in range(max(y.shape[0], x.shape[0])):
                    p = y[k, 0]
                    q = y[k, 1]
                    r = y[k, 2]
                    s = y[k, 3]

                    t = x[k, 0]
                    u = x[k, 1]
                    v = x[k, 2]
                    w = x[k, 3]

                    z[k, 0] = s * t + w * p + q * v - r * u
                    z[k, 1] = s * u + w * q + r * t - p * v
                    z[k, 2] = s * v + w * r + p * u - q * t
                    z[k, 3] = s * w - p * t - q * u - r * v

            x = z

    output = torch.empty(
        [x.shape[0], 3],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j in range(x.shape[0]):
        a = x[j, 0]
        b = x[j, 1]
        c = x[j, 2]
        d = x[j, 3]

        if d == 0 and (a == 0 and (b == 0 and c < 0 or b < 0) or a < 0) or d < 0:
            x[j] = -x[j]

        y = 2.0 * torch.atan2(torch.sqrt(a**2.0 + b**2.0 + c**2.0), d**1.0)

        if y < 0.001:
            y = 2.0 + y**2.0 / 12.0 + 7.0 * y**2.0 * y**2.0 / 2880.0
        else:
            y = y / torch.sin(y / 2.0)

        output[j] = x[j, :-1] * y

    if degrees:
        output = torch.rad2deg(output)

    return output



from pathlib import Path
from typing import Callable, Tuple, Union

from pandas import DataFrame

from beignet.transforms import Transform

from ._atom3d_dataset import ATOM3DDataset


class ATOM3DRESDataset(ATOM3DDataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = False,
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
    ):
        """
        ATOM3D Residue Identity (RES) consists of atomic environments
        extracted from non-redundant structures in the Protein Data Bank.
        This is formulated as a classification task where the identity of
        the amino acid in the center of the environment is predicted based
        on all other atoms.

        Each sample is a pair of features and a target, where features is
        the molecule’s atomic coordinates and target is the environments’s
        atomic coordinates

        Parameters
        ----------
        root : Union[str, Path]
            The root directory of the dataset.

        download : bool, optional
            If True, download the dataset from the specified source,
            by default `False`.

        transform : Union[Callable, Transform, None], optional
            The transformation function to be applied to the features,
            by default `None`.

        target_transform : Union[Callable, Transform, None], optional
            The transformation function to be applied to the targets,
            by default `None`.
        """
        super().__init__(
            root,
            "raw/RES/data",
            "https://zenodo.org/record/5026743/files/RES-raw.tar.gz",
            "RES",
            checksum="3d6b6c61efb890a8baa303280b6589d9",
            download=download,
        )

        self._transform_fn = transform

        self._target_transform_fn = target_transform

    def __getitem__(self, index: int) -> Tuple[DataFrame, DataFrame]:
        """
        Parameters
        ----------
        index : int
            The index of the item to retrieve from the dataset.

        Returns
        -------
        Tuple[DataFrame, DataFrame]
            A tuple containing the features and target of the item.
        """
        item = super().__getitem__(index)

        features = DataFrame(**item["atoms"])

        if self._transform_fn is not None:
            features = self._transform_fn(features)

        target = DataFrame(**item["labels"])

        if self._target_transform_fn is not None:
            target = self._target_transform_fn(target)

        return features, target



import beignet
import hypothesis.strategies
import torch.testing
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    canonical = function(hypothesis.strategies.booleans())

    rotations = Rotation.random(
        function(
            hypothesis.strategies.integers(
                min_value=1,
                max_value=8,
            ),
        ),
    )

    return (
        {
            "input": torch.from_numpy(rotations.as_quat(canonical)),
            "canonical": canonical,
        },
        torch.from_numpy(rotations.magnitude()),
    )


@hypothesis.given(_strategy())
def test_quaternion_magnitude(data):
    parameters, expected = data

    torch.testing.assert_close(
        beignet.quaternion_magnitude(**parameters),
        expected,
    )



import beignet
import torch


def test_probabilists_hermite_polynomial_domain():
    torch.testing.assert_close(
        beignet.probabilists_hermite_polynomial_domain,
        torch.tensor([-1.0, 1.0]),
        check_dtype=False,
    )



import math

import beignet
import torch


def test_evaluate_physicists_hermite_polynomial():
    coefficients = [
        torch.tensor([1]),
        torch.tensor([0, 2]),
        torch.tensor([-2, 0, 4]),
        torch.tensor([0, -12, 0, 8]),
        torch.tensor([12, 0, -48, 0, 16]),
        torch.tensor([0, 120, 0, -160, 0, 32]),
        torch.tensor([-120, 0, 720, 0, -480, 0, 64]),
        torch.tensor([0, -1680, 0, 3360, 0, -1344, 0, 128]),
        torch.tensor([1680, 0, -13440, 0, 13440, 0, -3584, 0, 256]),
        torch.tensor([0, 30240, 0, -80640, 0, 48384, 0, -9216, 0, 512]),
    ]

    output = beignet.evaluate_physicists_hermite_polynomial(
        torch.tensor([]),
        torch.tensor([1.0]),
    )

    assert math.prod(output.shape) == 0

    ys = []

    input = torch.linspace(-1, 1, 50)

    for coefficient in coefficients:
        ys = [
            *ys,
            beignet.evaluate_polynomial(
                input,
                coefficient,
            ),
        ]

    for index in range(10):
        torch.testing.assert_close(
            beignet.evaluate_physicists_hermite_polynomial(
                input,
                torch.tensor([0.0] * index + [1.0]),
            ),
            ys[index],
        )

    for index in range(3):
        shape = (2,) * index

        input = torch.zeros(shape)

        output = beignet.evaluate_physicists_hermite_polynomial(
            input,
            torch.tensor([1.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_physicists_hermite_polynomial(
            input,
            torch.tensor([1.0, 0.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_physicists_hermite_polynomial(
            input,
            torch.tensor([1.0, 0.0, 0.0]),
        )

        assert output.shape == shape



import torch
from torch import Tensor


def quaternion_identity(
    size: int,
    canonical: bool | None = False,
    *,
    out: Tensor | None = None,
    dtype: torch.dtype | None = None,
    layout: torch.layout | None = torch.strided,
    device: torch.device | None = None,
    requires_grad: bool | None = False,
) -> Tensor:
    r"""
    Identity rotation quaternions.

    Parameters
    ----------
    size : int
        Output size.

    canonical : bool, optional
        Whether to map the redundant double cover of rotation space to a unique
        canonical single cover. If `True`, then the rotation quaternion is
        chosen from :math:`{q, -q}` such that the :math:`w` term is positive.
        If the :math:`w` term is :math:`0`, then the rotation quaternion is
        chosen such that the first non-zero term of the :math:`x`, :math:`y`,
        and :math:`z` terms is positive.

    out : Tensor, optional
        Output tensor. Default, `None`.

    dtype : torch.dtype, optional
        Type of the returned tensor. Default, global default.

    layout : torch.layout, optional
        Layout of the returned tensor. Default, `torch.strided`.

    device : torch.device, optional
        Device of the returned tensor. Default, current device for the default
        tensor type.

    requires_grad : bool, optional
        Whether autograd records operations on the returned tensor. Default,
        `False`.

    Returns
    -------
    identity_quaternions : Tensor, shape (size, 4)
        Identity rotation quaternions.
    """
    rotation = torch.zeros(
        [size, 4],
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        requires_grad=requires_grad,
    )

    rotation[:, 3] = 1.0

    return rotation



import beignet
import torch


def test_probabilists_hermite_polynomial_weight():
    torch.testing.assert_close(
        beignet.probabilists_hermite_polynomial_weight(
            torch.linspace(-5, 5, 11),
        ),
        torch.exp(-0.5 * torch.linspace(-5, 5, 11) ** 2),
    )



import numpy
import torch

from ._evaluate_legendre_polynomial import evaluate_legendre_polynomial


def integrate_legendre_polynomial(
    input,
    order=1,
    k=None,
    lower_bound=0,
    scale=1,
    axis=0,
):
    if k is None:
        k = []

    input = torch.atleast_1d(input)

    lbnd, scl = map(torch.tensor, (lower_bound, scale))

    if not numpy.iterable(k):
        k = [k]

    if len(k) > order:
        raise ValueError("Too many integration constants")

    if lbnd.ndim != 0:
        raise ValueError("lbnd must be a scalar.")

    if scl.ndim != 0:
        raise ValueError("scl must be a scalar.")

    if order == 0:
        return input

    output = torch.moveaxis(input, axis, 0)

    k = torch.tensor(list(k) + [0] * (order - len(k)))
    k = torch.atleast_1d(k)

    for i in range(order):
        n = len(output)
        output *= scl
        tmp = torch.empty((n + 1,) + output.shape[1:], dtype=output.dtype)
        tmp[0] = output[0] * 0
        tmp[1] = output[0]
        if n > 1:
            tmp[2] = output[1] / 3

        if n < 2:
            j = torch.tensor([], dtype=torch.int32)
        else:
            j = torch.arange(2, n)

        t = (output[j].T / (2 * j + 1)).T
        tmp[j + 1] = t
        tmp[j - 1] += -t
        legval_value = evaluate_legendre_polynomial(lbnd, tmp)
        tmp[0] += k[i] - legval_value
        output = tmp

    output = torch.moveaxis(output, 0, axis)

    return output



import torch
from torch import Tensor


def subtract_polynomial(input: Tensor, other: Tensor) -> Tensor:
    r"""
    Returns the difference of two polynomials.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    other : Tensor
        Polynomial coefficients.

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the difference.
    """
    input = torch.atleast_1d(input)
    other = torch.atleast_1d(other)

    dtype = torch.promote_types(input.dtype, other.dtype)

    input = input.to(dtype)
    other = other.to(dtype)

    if input.shape[0] > other.shape[0]:
        output = -other

        output = torch.concatenate(
            [
                output,
                torch.zeros(
                    input.shape[0] - other.shape[0],
                    dtype=other.dtype,
                ),
            ],
        )
        output = input + output
    else:
        output = -other

        output = torch.concatenate(
            [
                output[: input.shape[0]] + input,
                output[input.shape[0] :],
            ],
        )

    return output



import pytest
import torch
from beignet.features import RotationVector
from torch import Size


class TestRotationVector:
    @pytest.fixture
    def rotation_vector(self) -> RotationVector:
        return RotationVector(torch.tensor([[1, 2, 3]]))

    def test___new__(self):
        result = RotationVector(torch.tensor([[1, 2, 3]]))

        assert isinstance(result, RotationVector)

        assert result.shape == Size([1, 3])

        with pytest.raises(ValueError):
            RotationVector(torch.tensor(5))

        with pytest.raises(ValueError):
            RotationVector(torch.tensor([1, 2]))

    def test___repr__(self, rotation_vector):
        assert isinstance(rotation_vector.__repr__(), str)

    def test__wrap(self):
        result = RotationVector._wrap(torch.tensor([[1, 2, 3]]))

        assert isinstance(result, RotationVector)

        assert result.shape == Size([1, 3])

    def test_wrap_like(self, rotation_vector):
        result = RotationVector.wrap_like(
            rotation_vector,
            torch.tensor([[1, 2, 3]]),
        )

        assert isinstance(result, RotationVector)

        assert result.shape == Size([1, 3])



import math
from typing import Literal

import torch
import torch.nn.functional
from torch import Tensor


def convolve(
    input: Tensor,
    other: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    if input.ndim != other.ndim:
        raise ValueError

    for i in range(input.ndim - 1):
        a = input.size(i)
        b = other.size(i)

        if a == b or a == 1 or b == 1:
            continue

        raise ValueError

    if mode not in {"full", "same", "valid"}:
        raise ValueError

    x_size, y_size = input.shape[-1], other.shape[-1]

    if input.shape[-1] < other.shape[-1]:
        input, other = other, input

    if input.shape[:-1] != other.shape[:-1]:
        input_shape = []

        for i, j in zip(input.shape[:-1], other.shape[:-1], strict=False):
            input_shape = [*input_shape, max(i, j)]

        input = torch.broadcast_to(
            input,
            [*input_shape, input.shape[-1]],
        )

        other = torch.broadcast_to(
            other,
            [*input_shape, other.shape[-1]],
        )

    input = torch.reshape(
        input,
        [math.prod(input.shape[:-1]), input.shape[-1]],
    )

    other = torch.unsqueeze(
        torch.flip(
            torch.reshape(
                other,
                [math.prod(input.shape[:-1]), other.shape[-1]],
            ),
            dims=[-1],
        ),
        dim=1,
    )

    output = torch.reshape(
        torch.nn.functional.conv1d(
            input,
            other,
            groups=input.shape[0],
            padding=other.shape[-1] - 1,
        ),
        [*input.shape[:-1], -1],
    )

    match mode:
        case "same":
            size = x_size

            m = (output.shape[-1] - size) // 2
            n = m + size

            output = output[..., m:n]
        case "valid":
            size = max(x_size, y_size) - min(x_size, y_size) + 1

            m = (output.shape[-1] - size) // 2
            n = m + size

            output = output[..., m:n]

    return output



import torch
from torch import Tensor


def evaluate_polynomial_from_roots(
    input: Tensor,
    other: Tensor,
    tensor: bool = True,
) -> Tensor:
    if other.ndim == 0:
        other = torch.ravel(other)

    if tensor:
        other = torch.reshape(other, other.shape + (1,) * input.ndim)

    if input.ndim >= other.ndim:
        raise ValueError

    output = torch.prod(input - other, dim=0)

    return output



import beignet
import torch


def test_physicists_hermite_polynomial_from_roots():
    torch.testing.assert_close(
        beignet.trim_physicists_hermite_polynomial_coefficients(
            beignet.physicists_hermite_polynomial_from_roots(
                torch.tensor([]),
            ),
            tol=0.000001,
        ),
        torch.tensor([1.0]),
    )

    # for i in range(1, 5):
    #     roots = torch.cos(torch.linspace(-math.pi, 0, 2 * i + 1)[1::2])
    #     target = 0
    #
    #     torch.testing.assert_close(
    #         beignet.herm2poly(
    #             beignet.hermfromroots(
    #                 roots,
    #             ),
    #         )[-1],
    #         torch.tensor([1.0]),
    #     )
    #
    #     torch.testing.assert_close(
    #         beignet.hermval(
    #             roots,
    #             beignet.hermfromroots(
    #                 roots,
    #             ),
    #         ),
    #         target,
    #     )



import beignet
import pytest
import torch


def test_evaluate_physicists_hermite_polynomial_2d():
    input = torch.rand(3, 5) * 2 - 1

    a, b, c = input

    x, y, z = beignet.evaluate_polynomial(
        input,
        torch.tensor([1.0, 2.0, 3.0]),
    )

    with pytest.raises(ValueError):
        beignet.evaluate_physicists_hermite_polynomial_2d(
            a,
            b[:2],
            torch.einsum(
                "i,j->ij",
                torch.tensor([2.5, 1.0, 0.75]),
                torch.tensor([2.5, 1.0, 0.75]),
            ),
        )

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial_2d(
            a,
            b,
            torch.einsum(
                "i,j->ij",
                torch.tensor([2.5, 1.0, 0.75]),
                torch.tensor([2.5, 1.0, 0.75]),
            ),
        ),
        x * y,
    )

    output = beignet.evaluate_physicists_hermite_polynomial_2d(
        torch.ones([2, 3]),
        torch.ones([2, 3]),
        torch.einsum(
            "i,j->ij",
            torch.tensor([2.5, 1.0, 0.75]),
            torch.tensor([2.5, 1.0, 0.75]),
        ),
    )

    assert output.shape == (2, 3)



import beignet
import hypothesis.strategies
from beignet.datasets import RandomRotationMatrixDataset


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=1,
            max_value=8,
        ),
    )

    return (
        {
            "size": size,
        },
        beignet.random_rotation_matrix(size),
    )


class TestRandomRotationMatrixDataset:
    @hypothesis.given(_strategy())
    def test___init__(self, data):
        parameters, output = data

        dataset = RandomRotationMatrixDataset(**parameters)

        assert dataset.data.shape == output.shape

        assert dataset.data.dtype == output.dtype

        assert dataset.data.layout == output.layout



from pathlib import Path
from typing import Callable, Optional, Union

import beignet.io
from beignet.transforms import Transform

from ._lmdb_dataset import LMDBDataset


class ATOM3DDataset(LMDBDataset):
    def __init__(
        self,
        root: Union[str, Path],
        path: Union[str, Path],
        resource: str,
        name: str,
        *,
        checksum: Optional[str] = None,
        download: bool = False,
        transform: Union[Callable, Transform, None] = None,
    ):
        self._root = root

        if isinstance(self._root, str):
            self._root = Path(self._root).resolve()

        self._transform_fn = transform

        if download:
            beignet.io.download_and_extract_archive(
                resource,
                self._root / f"ATOM3D{name}",
                checksum=checksum,
            )

        super().__init__(
            self._root / f"ATOM3D{name}" / path,
            transform=transform,
        )



import torch
from torch import Tensor


def trim_laguerre_polynomial_coefficients(
    input: Tensor,
    tol: float = 0.0,
) -> Tensor:
    if tol < 0:
        raise ValueError

    input = torch.atleast_1d(input)

    indices = torch.nonzero(torch.abs(input) > tol)

    if indices.shape[0] == 0:
        output = input[:1] * 0
    else:
        output = input[: indices[-1] + 1]

    return output



from typing import Any, Callable, Dict, Type

from ._transform import Transform


class Lambda(Transform):
    _transformed_types = (object,)

    def __init__(self, fn: Callable[[Any], Any], *types: Type):
        """
        Parameters
        ----------

        fn : Callable[[Any], Any]
            The function to be used as the transformation function.

        types : Type
            The types of the input arguments that will be passed to the
            function. If no types are provided, the default transformed types
            will be used.
        """
        super().__init__()

        self._fn = fn

        self._types = types or self._transformed_types

    def _transform(self, input: Any, parameters: Dict[str, Any]) -> Any:
        """
        Parameters
        ----------

        input : Any
            The input value to be transformed.

        parameters : Dict[str, Any]
            A dictionary containing any additional parameters required for the
            transformation.

        Returns
        -------
        Any
            The transformed value.

        """
        if isinstance(input, self._types):
            return self._fn(input)
        else:
            return input

    def extra_repr(self) -> str:
        """
        Get a string representation of the Lambda transform.

        Returns
        -------
        str
            A string representation of the Lambda transform, including the
            function name and types.
        """
        extras = []

        name = getattr(self._fn, "__name__", None)

        if name:
            extras.append(name)

        extras.append(f"types={[type.__name__ for type in self._types]}")

        return ", ".join(extras)



import torch
from torch import Tensor

from ._add_laguerre_polynomial import add_laguerre_polynomial
from ._multiply_laguerre_polynomial_by_x import multiply_laguerre_polynomial_by_x


def polynomial_to_laguerre_polynomial(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    output = torch.zeros_like(input)

    for i in range(0, input.shape[0]):
        output = multiply_laguerre_polynomial_by_x(output, mode="same")

        output = add_laguerre_polynomial(output, torch.flip(input, dims=[0])[i])

    return output



import beignet
import pytest
import torch


def test_differentiate_chebyshev_polynomial():
    with pytest.raises(TypeError):
        beignet.differentiate_chebyshev_polynomial(
            torch.tensor([0.0]),
            torch.tensor([0.5]),
        )

    with pytest.raises(ValueError):
        beignet.differentiate_chebyshev_polynomial(
            torch.tensor([0.0]),
            order=torch.tensor([-1.0]),
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.trim_chebyshev_polynomial_coefficients(
                beignet.differentiate_chebyshev_polynomial(
                    torch.tensor([0.0] * i + [1.0]),
                    order=torch.tensor([0.0]),
                ),
                tol=0.000001,
            ),
            beignet.trim_chebyshev_polynomial_coefficients(
                torch.tensor([0.0] * i + [1.0]),
                tol=0.000001,
            ),
        )

    # for i in range(5):
    #     for j in range(2, 5):
    #         torch.testing.assert_close(
    #             beignet.chebtrim(
    #                 beignet.chebder(
    #                     beignet.chebint(
    #                         torch.tensor([0.0] * i + [1.0]), order=j
    #                     ),
    #                     order=j,
    #                 ),
    #                 tol=0.000001,
    #             ),
    #             beignet.chebtrim(
    #                 torch.tensor([0.0] * i + [1.0]),
    #                 tol=0.000001,
    #             ),
    #         )

    # for i in range(5):
    #     for j in range(2, 5):
    #         torch.testing.assert_close(
    #             beignet.chebtrim(
    #                 beignet.chebder(
    #                     beignet.chebint(
    #                         torch.tensor([0.0] * i + [1.0]),
    #                         order=j,
    #                         scale=2.0,
    #                     ),
    #                     order=j,
    #                     scale=0.5,
    #                 ),
    #                 tol=0.000001,
    #             ),
    #             beignet.chebtrim(
    #                 torch.tensor([0.0] * i + [1.0]),
    #                 tol=0.000001,
    #             ),
    #         )

    input = torch.rand(3, 4)

    torch.testing.assert_close(
        beignet.differentiate_chebyshev_polynomial(
            input,
            axis=0,
        ),
        torch.vstack(
            [beignet.differentiate_chebyshev_polynomial(c) for c in input.T]
        ).T,
    )

    torch.testing.assert_close(
        beignet.differentiate_chebyshev_polynomial(
            input,
            axis=1,
        ),
        torch.vstack([beignet.differentiate_chebyshev_polynomial(c) for c in input]),
    )



import math

import torch
from torch import Tensor


def chebyshev_zeros(input: int) -> Tensor:
    if input < 1:
        raise ValueError

    return torch.sin(0.5 * math.pi / input * torch.arange(-input + 1, input + 1, 2))



from os import PathLike
from typing import Any, Callable

try:
    from mdtraj import Trajectory
except ImportError:
    pass
import beignet

from ._trajectory_dataset import TrajectoryDataset


class HDF5TrajectoryDataset(TrajectoryDataset):
    def __init__(
        self,
        root: str | PathLike,
        transform: Callable[["Trajectory"], Any] | None = None,
        stride: int | None = None,
        **kwargs,
    ):
        if beignet.optional_dependencies(["mdtraj"], ["mdtraj"]):
            import mdtraj

            super().__init__(
                func=mdtraj.load_hdf5,
                extension="hdf5",
                root=root,
                transform=transform,
                stride=stride,
                **kwargs,
            )



from typing import Callable, TypeVar

import torch
from torch import Tensor

import beignet

T = TypeVar("T")


def space(
    box: Tensor | None = None,
    *,
    normalized: bool = True,
    parallelepiped: bool = True,
    remapped: bool = True,
) -> (Callable, Callable):
    r"""Define a simulation space.

    This function is fundamental in constructing simulation spaces derived from
    subsets of $\mathbb{R}^{D}$ (where $D = 1$, $2$, or $3$) and is
    instrumental in setting up simulation environments with specific
    characteristics (e.g., periodic boundary conditions). The function returns
    a displacement function and a shift function to compute particle
    interactions and movements in space.

    This function supports deformation of the simulation cell, crucial for
    certain types of simulations, such as those involving finite deformations
    or the computation of elastic constants.

    Parameters
    ----------
    box : Tensor | None, default=None
        Interpretation varies based on the value of `parallelepiped`. If
        `parallelepiped` is `True`, must be an affine transformation, $T$,
        specified in one of three ways:

        1. a cube, $L$;
        2. an orthorhombic unit cell, $[L_{x}, L_{y}, L_{z}]$; or
        3. a triclinic cell, upper triangular matrix.

        If `parallelepiped` is `False`, must be the edge lengths.

        If `box` is `None`, the simulation space has free boundary conditions.

    normalized : bool, default=True
        If `normalized` is `True`, positions are stored in the unit cube.
        Displacements and shifts are computed in a normalized simulation space
        and can be transformed back to real simulation space using the affine
        transformation. If `normalized` is `False`, positions are expressed and
        computations performed directly in the real simulation space.

    parallelepiped : bool, default=True
        If `True`, the simulation space is defined as a ${1, 2, 3}$-dimensional
        parallelepiped with periodic boundary conditions. If `False`, the space
        is defined on a ${1, 2, 3}$-dimensional hypercube.

    remapped : bool, default=True
        If `True`, positions and displacements are remapped to stay in the
        bounds of the defined simulation space. A rempapped simulation space is
        topologically equivalent to a torus, ensuring that particles exiting
        one boundary re-enter from the opposite side.

    Returns
    -------
    (Callable[[Tensor, Tensor], Tensor], Callable[[Tensor, Tensor], Tensor])
        A pair of functions:

        1.  The displacement function, $\vec{d}$, measures the
            difference between two points in the simulation space, factoring in
            the geometry and boundary conditions. This function is used to
            calculate particle interactions and dynamics.

        2.  The shift function, $u$, applies a displacement vector to a point
            in the space, effectively moving it. This function is used to
            update simulated particle positions.

    Examples
    --------
        transformation = torch.tensor([10.0])

        displacement_fn, shift_fn = space(
            transformation,
            normalized=False,
        )

        normalized_displacement_fn, normalized_shift_fn = space(
            transformation,
            normalized=True,
        )

        normalized_position = torch.rand([4, 3])

        position = transformation * normalized_position

        displacement = torch.randn([4, 3])

        torch.testing.assert_close(
            displacement_fn(position[0], position[1]),
            normalized_displacement_fn(
                normalized_position[0],
                normalized_position[1],
            ),
        )
    """
    if isinstance(box, (int, float)):
        box = torch.tensor([box])

    if box is None:

        def displacement_fn(
            input: Tensor,
            other: Tensor,
            *,
            perturbation: Tensor | None = None,
            **_,
        ) -> Tensor:
            if len(input.shape) != 1:
                raise ValueError

            if input.shape != other.shape:
                raise ValueError

            if perturbation is not None:
                transform = input - other

                match transform.ndim:
                    case 0:
                        return perturbation * transform
                    case 1:
                        return torch.einsum(
                            "i,...i->...i",
                            transform,
                            perturbation,
                        )
                    case 2:
                        return torch.einsum(
                            "ij,...j->...i",
                            transform,
                            perturbation,
                        )
                    case _:
                        raise ValueError

            return input - other

        def shift_fn(input: Tensor, other: Tensor, **_) -> Tensor:
            return input + other

        return displacement_fn, shift_fn

    if parallelepiped:
        inverted_transform = beignet.invert_transform(box)

        if normalized:

            def displacement_fn(
                input: Tensor,
                other: Tensor,
                *,
                perturbation: Tensor | None = None,
                **kwargs,
            ) -> Tensor:
                _transform = box

                _inverted_transform = inverted_transform

                if "transform" in kwargs:
                    _transform = kwargs["transform"]

                if "updated_transform" in kwargs:
                    _transform = kwargs["updated_transform"]

                if len(input.shape) != 1:
                    raise ValueError

                if input.shape != other.shape:
                    raise ValueError

                displacement = beignet.apply_transform(
                    torch.remainder(input - other + 1.0 * 0.5, 1.0) - 1.0 * 0.5,
                    _transform,
                )

                if perturbation is not None:
                    match displacement.ndim:
                        case 0:
                            return perturbation * displacement
                        case 1:
                            return torch.einsum(
                                "i,...i->...i",
                                displacement,
                                perturbation,
                            )
                        case 2:
                            return torch.einsum(
                                "ij,...j->...i",
                                displacement,
                                perturbation,
                            )
                        case _:
                            raise ValueError

                return displacement

            if remapped:

                def u(input: Tensor, other: Tensor) -> Tensor:
                    return torch.remainder(input + other, 1.0)

                def shift_fn(input: Tensor, other: Tensor, **kwargs) -> Tensor:
                    _transform = box

                    _inverted_transform = inverted_transform

                    if "transform" in kwargs:
                        _transform = kwargs["transform"]

                        _inverted_transform = beignet.invert_transform(_transform)

                    if "updated_transform" in kwargs:
                        _transform = kwargs["updated_transform"]

                    return u(input, beignet.apply_transform(other, _inverted_transform))

                return displacement_fn, shift_fn

            def shift_fn(input: Tensor, other: Tensor, **kwargs) -> Tensor:
                _transform = box

                _inverted_transform = inverted_transform

                if "transform" in kwargs:
                    _transform = kwargs["transform"]

                    _inverted_transform = beignet.invert_transform(_transform)

                if "updated_transform" in kwargs:
                    _transform = kwargs["updated_transform"]

                return input + beignet.apply_transform(other, _inverted_transform)

            return displacement_fn, shift_fn

        def displacement_fn(
            input: Tensor,
            other: Tensor,
            *,
            perturbation: Tensor | None = None,
            **kwargs,
        ) -> Tensor:
            _transform = box

            _inverted_transform = inverted_transform

            if "transform" in kwargs:
                _transform = kwargs["transform"]

                _inverted_transform = beignet.invert_transform(_transform)

            if "updated_transform" in kwargs:
                _transform = kwargs["updated_transform"]

            input = beignet.apply_transform(input, _inverted_transform)
            other = beignet.apply_transform(other, _inverted_transform)

            if len(input.shape) != 1:
                raise ValueError

            if input.shape != other.shape:
                raise ValueError

            displacement = beignet.apply_transform(
                torch.remainder(input - other + 1.0 * 0.5, 1.0) - 1.0 * 0.5,
                _transform,
            )

            if perturbation is not None:
                match displacement.ndim:
                    case 0:
                        return perturbation * displacement
                    case 1:
                        return torch.einsum(
                            "i,...i->...i",
                            displacement,
                            perturbation,
                        )
                    case 2:
                        return torch.einsum(
                            "ij,...j->...i",
                            displacement,
                            perturbation,
                        )
                    case _:
                        raise ValueError

            return displacement

        if remapped:

            def u(input: Tensor, other: Tensor) -> Tensor:
                return torch.remainder(input + other, 1.0)

            def shift_fn(input: Tensor, other: Tensor, **kwargs) -> Tensor:
                _transform = box

                _inverted_transform = inverted_transform

                if "transform" in kwargs:
                    _transform = kwargs["transform"]

                    _inverted_transform = beignet.invert_transform(
                        _transform,
                    )

                if "updated_transform" in kwargs:
                    _transform = kwargs["updated_transform"]

                return beignet.apply_transform(
                    u(
                        beignet.apply_transform(_inverted_transform, input),
                        beignet.apply_transform(_inverted_transform, other),
                    ),
                    _transform,
                )

            return displacement_fn, shift_fn

        def shift_fn(input: Tensor, other: Tensor, **_) -> Tensor:
            return input + other

        return displacement_fn, shift_fn

    def displacement_fn(
        input: Tensor,
        other: Tensor,
        *,
        perturbation: Tensor | None = None,
        **_,
    ) -> Tensor:
        if len(input.shape) != 1:
            raise ValueError

        if input.shape != other.shape:
            raise ValueError

        displacement = torch.remainder(
            input - other + box * 0.5,
            box,
        )

        if perturbation is not None:
            transform = displacement - box * 0.5

            match transform.ndim:
                case 0:
                    return perturbation * transform
                case 1:
                    return torch.einsum(
                        "i,...i->...i",
                        transform,
                        perturbation,
                    )
                case 2:
                    return torch.einsum(
                        "ij,...j->...i",
                        transform,
                        perturbation,
                    )
                case _:
                    raise ValueError

        return displacement - box * 0.5

    if remapped:

        def shift_fn(input: Tensor, other: Tensor, **_) -> Tensor:
            return torch.remainder(input + other, box)
    else:

        def shift_fn(input: Tensor, other: Tensor, **_) -> Tensor:
            return input + other

    return displacement_fn, shift_fn



import beignet
import pytest
import torch


def test_integrate_physicists_hermite_polynomial():
    with pytest.raises(TypeError):
        beignet.integrate_physicists_hermite_polynomial(
            torch.tensor([0.0]),
            order=0.5,
        )

    with pytest.raises(ValueError):
        beignet.integrate_physicists_hermite_polynomial(
            torch.tensor([0]),
            order=-1,
        )

    with pytest.raises(ValueError):
        beignet.integrate_physicists_hermite_polynomial(
            torch.tensor([0]),
            order=1,
            k=torch.tensor([0, 0]),
        )

    with pytest.raises(ValueError):
        beignet.integrate_physicists_hermite_polynomial(
            torch.tensor([0]),
            lower_bound=[0],
        )

    with pytest.raises(ValueError):
        beignet.integrate_physicists_hermite_polynomial(
            torch.tensor([0]),
            scale=[0],
        )

    with pytest.raises(TypeError):
        beignet.integrate_physicists_hermite_polynomial(
            torch.tensor([0]),
            axis=0.5,
        )

    for i in range(2, 5):
        torch.testing.assert_close(
            beignet.trim_physicists_hermite_polynomial_coefficients(
                beignet.integrate_physicists_hermite_polynomial(
                    torch.tensor([0.0]),
                    order=i,
                    k=([0.0] * (i - 2) + [1.0]),
                ),
                tol=0.000001,
            ),
            torch.tensor([0.0, 0.5]),
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.trim_physicists_hermite_polynomial_coefficients(
                beignet.physicists_hermite_polynomial_to_polynomial(
                    beignet.integrate_physicists_hermite_polynomial(
                        beignet.polynomial_to_physicists_hermite_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                        ),
                        order=1,
                        k=[i],
                    )
                ),
                tol=0.000001,
            ),
            beignet.trim_physicists_hermite_polynomial_coefficients(
                torch.tensor([i] + [0.0] * i + [1.0 / (i + 1)]),
                tol=0.000001,
            ),
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.evaluate_physicists_hermite_polynomial(
                torch.tensor([-1.0]),
                beignet.integrate_physicists_hermite_polynomial(
                    beignet.polynomial_to_physicists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                    ),
                    order=1,
                    k=[i],
                    lower_bound=-1,
                ),
            ),
            torch.tensor([i], dtype=torch.get_default_dtype()),
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.trim_physicists_hermite_polynomial_coefficients(
                beignet.physicists_hermite_polynomial_to_polynomial(
                    beignet.integrate_physicists_hermite_polynomial(
                        beignet.polynomial_to_physicists_hermite_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                        ),
                        order=1,
                        k=[i],
                        scale=2,
                    ),
                ),
                tol=0.000001,
            ),
            beignet.trim_physicists_hermite_polynomial_coefficients(
                torch.tensor([i] + [0.0] * i + [2.0 / (i + 1.0)]),
                tol=0.000001,
            ),
        )

    for i in range(5):
        for j in range(2, 5):
            target = torch.tensor([0.0] * i + [1.0])[:]

            for _ in range(j):
                target = beignet.integrate_physicists_hermite_polynomial(
                    target,
                    order=1,
                )

            torch.testing.assert_close(
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    beignet.integrate_physicists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                        order=j,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            pol = torch.tensor([0.0] * i + [1.0])

            target = pol[:]

            for k in range(j):
                target = beignet.integrate_physicists_hermite_polynomial(
                    target, order=1, k=[k]
                )

            torch.testing.assert_close(
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    beignet.integrate_physicists_hermite_polynomial(
                        pol,
                        order=j,
                        k=list(range(j)),
                    ),
                    tol=0.000001,
                ),
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            pol = torch.tensor([0.0] * i + [1.0])
            target = pol[:]
            for k in range(j):
                target = beignet.integrate_physicists_hermite_polynomial(
                    target,
                    order=1,
                    k=[k],
                    lower_bound=-1,
                )

            torch.testing.assert_close(
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    beignet.integrate_physicists_hermite_polynomial(
                        pol,
                        order=j,
                        k=list(range(j)),
                        lower_bound=-1,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            pol = torch.tensor([0.0] * i + [1.0])
            target = pol[:]
            for k in range(j):
                target = beignet.integrate_physicists_hermite_polynomial(
                    target,
                    order=1,
                    k=[k],
                    scale=2,
                )

            # torch.testing.assert_close(
            #     beignet.hermtrim(
            #         beignet.hermint(
            #             pol,
            #             order=j,
            #             k=list(range(j)),
            #             scale=2,
            #         ),
            #         tol=0.000001,
            #     ),
            #     beignet.hermtrim(
            #         target,
            #         tol=0.000001,
            #     ),
            # )

    c2d = torch.rand(3, 4)

    target = torch.vstack(
        [beignet.integrate_physicists_hermite_polynomial(c) for c in c2d.T]
    ).T

    torch.testing.assert_close(
        beignet.integrate_physicists_hermite_polynomial(
            c2d,
            axis=0,
        ),
        target,
    )

    target = torch.vstack(
        [beignet.integrate_physicists_hermite_polynomial(c) for c in c2d]
    )

    torch.testing.assert_close(
        beignet.integrate_physicists_hermite_polynomial(
            c2d,
            axis=1,
        ),
        target,
    )

    target = torch.vstack(
        [beignet.integrate_physicists_hermite_polynomial(c, k=3) for c in c2d]
    )

    torch.testing.assert_close(
        beignet.integrate_physicists_hermite_polynomial(
            c2d,
            k=3,
            axis=1,
        ),
        target,
    )



from torch import Tensor

from ._quaternion_magnitude import quaternion_magnitude
from ._rotation_vector_to_quaternion import (
    rotation_vector_to_quaternion,
)


def rotation_vector_magnitude(
    input: Tensor,
    degrees: bool | None = False,
) -> Tensor:
    r"""
    Rotation vector magnitudes.

    Parameters
    ----------
    input : Tensor, shape (..., 3)
        Rotation vectors.

    degrees : bool, optional
        If `True`, magnitudes are assumed to be in degrees. Default, `False`.

    Returns
    -------
    rotation_vector_magnitudes : Tensor, shape (...)
        Angles in radians. Magnitudes will be in the range :math:`[0, \pi]`.
    """
    return quaternion_magnitude(
        rotation_vector_to_quaternion(
            input,
            degrees,
        ),
    )



import torch
from torch import Tensor


def chebyshev_polynomial_vandermonde(
    x: Tensor,
    degree: Tensor,
) -> Tensor:
    if degree < 0:
        raise ValueError

    x = torch.atleast_1d(x)
    dims = (degree + 1,) + x.shape
    dtyp = torch.promote_types(x.dtype, torch.tensor(0.0).dtype)
    x = x.to(dtyp)
    v = torch.empty(dims, dtype=dtyp)

    v[0] = torch.ones_like(x)

    if degree > 0:
        v[1] = x

        x2 = 2 * x

        for index in range(2, degree + 1):
            v[index] = v[index - 1] * x2 - v[index - 2]

    return torch.moveaxis(v, 0, -1)



from ._dawson_integral_f import dawson_integral_f
from ._error_erf import error_erf
from ._error_erfc import error_erfc
from ._error_erfi import error_erfi
from ._faddeeva_w import faddeeva_w

__all__ = [
    "dawson_integral_f",
    "error_erf",
    "error_erfc",
    "error_erfi",
    "faddeeva_w",
]



import math

import beignet
import torch


# def test_chebyshev_polynomial_from_roots():
#     torch.testing.assert_close(
#         beignet.trim_chebyshev_polynomial_coefficients(
#             beignet.chebyshev_polynomial_from_roots(
#                 torch.tensor([]),
#             ),
#             tol=0.000001,
#         ),
#         torch.tensor([1.0]),
#     )
#
#     for index in range(1, 5):
#         input = beignet.chebyshev_polynomial_from_roots(
#             torch.cos(torch.linspace(-math.pi, 0.0, 2 * index + 1)[1::2]),
#         )
#
#         input = input * 2 ** (index - 1)
#
#         torch.testing.assert_close(
#             beignet.trim_chebyshev_polynomial_coefficients(
#                 input,
#                 tol=0.000001,
#             ),
#             beignet.trim_chebyshev_polynomial_coefficients(
#                 torch.tensor([0.0] * index + [1.0]),
#                 tol=0.000001,
#             ),
#         )



import torch
from torch import Tensor

from ._quaternion_identity import quaternion_identity
from ._quaternion_to_rotation_matrix import (
    quaternion_to_rotation_matrix,
)


def rotation_matrix_identity(
    size: int,
    *,
    out: Tensor | None = None,
    dtype: torch.dtype | None = None,
    layout: torch.layout | None = torch.strided,
    device: torch.device | None = None,
    requires_grad: bool | None = False,
) -> Tensor:
    r"""
    Identity rotation matrices.

    Parameters
    ----------
    size : int
        Output size.

    out : Tensor, optional
        Output tensor. Default, `None`.

    dtype : torch.dtype, optional
        Type of the returned tensor. Default, global default.

    layout : torch.layout, optional
        Layout of the returned tensor. Default, `torch.strided`.

    device : torch.device, optional
        Device of the returned tensor. Default, current device for the default
        tensor type.

    requires_grad : bool, optional
        Whether autograd records operations on the returned tensor. Default,
        `False`.

    Returns
    -------
    identity_rotation_matrices : Tensor, shape (size, 3, 3)
        Identity rotation matrices.
    """
    return quaternion_to_rotation_matrix(
        quaternion_identity(
            size,
            out=out,
            dtype=dtype,
            layout=layout,
            device=device,
            requires_grad=requires_grad,
        ),
    )



import beignet
import pytest
import torch


def test_differentiate_physicists_hermite_polynomial():
    with pytest.raises(TypeError):
        beignet.differentiate_physicists_hermite_polynomial(torch.tensor([0]), 0.5)

    with pytest.raises(ValueError):
        beignet.differentiate_physicists_hermite_polynomial(torch.tensor([0]), -1)

    for i in range(5):
        torch.testing.assert_close(
            beignet.trim_physicists_hermite_polynomial_coefficients(
                beignet.differentiate_physicists_hermite_polynomial(
                    torch.tensor([0.0] * i + [1.0]),
                    order=0,
                ),
                tol=0.000001,
            ),
            beignet.trim_physicists_hermite_polynomial_coefficients(
                torch.tensor([0.0] * i + [1.0]),
                tol=0.000001,
            ),
        )

    for i in range(5):
        for j in range(2, 5):
            torch.testing.assert_close(
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    beignet.differentiate_physicists_hermite_polynomial(
                        beignet.integrate_physicists_hermite_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                            order=j,
                        ),
                        order=j,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    torch.tensor([0.0] * i + [1.0]),
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            torch.testing.assert_close(
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    beignet.differentiate_physicists_hermite_polynomial(
                        beignet.integrate_physicists_hermite_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                            order=j,
                            scale=2,
                        ),
                        order=j,
                        scale=0.5,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_physicists_hermite_polynomial_coefficients(
                    torch.tensor([0.0] * i + [1.0]),
                    tol=0.000001,
                ),
            )

    c2d = torch.rand(3, 4)

    torch.testing.assert_close(
        beignet.differentiate_physicists_hermite_polynomial(c2d, axis=0),
        torch.vstack(
            [beignet.differentiate_physicists_hermite_polynomial(c) for c in c2d.T]
        ).T,
    )

    torch.testing.assert_close(
        beignet.differentiate_physicists_hermite_polynomial(
            c2d,
            axis=1,
        ),
        torch.vstack(
            [beignet.differentiate_physicists_hermite_polynomial(c) for c in c2d]
        ),
    )



import beignet
import hypothesis.strategies
import torch
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    rotation = Rotation.random(
        function(
            hypothesis.strategies.integers(
                min_value=16,
                max_value=32,
            ),
        ),
    )

    degrees = function(hypothesis.strategies.booleans())

    return (
        {
            "input": torch.from_numpy(
                rotation.as_quat(
                    canonical=False,
                ),
            ),
            "degrees": degrees,
        },
        torch.from_numpy(
            rotation.as_rotvec(
                degrees,
            ),
        ),
    )


@hypothesis.given(_strategy())
def test_quaternion_to_rotation_vector(data):
    parameters, expected = data

    torch.testing.assert_close(
        beignet.quaternion_to_rotation_vector(
            **parameters,
        ),
        expected,
    )



from torch import Tensor

from ._quaternion_mean import quaternion_mean
from ._quaternion_to_rotation_matrix import quaternion_to_rotation_matrix
from ._rotation_matrix_to_quaternion import rotation_matrix_to_quaternion


def rotation_matrix_mean(
    input: Tensor,
    weight: Tensor | None = None,
) -> Tensor:
    r"""
    Parameters
    ----------
    input : Tensor, shape=(..., 3, 3)
        Rotation matrices.

    weight : Tensor, shape=(..., 4), optional
        Relative importance of rotation matrices.

    Returns
    -------
    output : Tensor, shape=(..., 3, 3)
    """
    return quaternion_to_rotation_matrix(
        quaternion_mean(
            rotation_matrix_to_quaternion(
                input,
            ),
            weight,
        ),
    )



import torch
from torch import Generator, Tensor


def random_quaternion(
    size: int,
    canonical: bool = False,
    *,
    generator: Generator | None = None,
    out: Tensor | None = None,
    dtype: torch.dtype | None = None,
    layout: torch.layout | None = torch.strided,
    device: torch.device | None = None,
    requires_grad: bool | None = False,
    pin_memory: bool | None = False,
) -> Tensor:
    r"""
    Generate random rotation quaternions.

    Parameters
    ----------
    size : int
        Output size.

    canonical : bool, optional
        Whether to map the redundant double cover of rotation space to a unique
        canonical single cover. If `True`, then the rotation quaternion is
        chosen from :math:`{q, -q}` such that the :math:`w` term is positive.
        If the :math:`w` term is :math:`0`, then the rotation quaternion is
        chosen such that the first non-zero term of the :math:`x`, :math:`y`,
        and :math:`z` terms is positive.

    generator : torch.Generator, optional
        Psuedo-random number generator. Default, `None`.

    out : Tensor, optional
        Output tensor. Default, `None`.

    dtype : torch.dtype, optional
        Type of the returned tensor. Default, global default.

    layout : torch.layout, optional
        Layout of the returned tensor. Default, `torch.strided`.

    device : torch.device, optional
        Device of the returned tensor. Default, current device for the default
        tensor type.

    requires_grad : bool, optional
        Whether autograd records operations on the returned tensor. Default,
        `False`.

    pin_memory : bool, optional
        If `True`, returned tensor is allocated in pinned memory. Default,
        `False`.

    Returns
    -------
    random_quaternions : Tensor, shape (..., 4)
        Random rotation quaternions.
    """
    quaternions = torch.rand(
        [size, 4],
        generator=generator,
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        requires_grad=requires_grad,
        pin_memory=pin_memory,
    )

    if canonical:
        for index in range(quaternions.size(0)):
            if (
                (quaternions[index][3] < 0)
                or (quaternions[index][3] == 0 and quaternions[index][0] < 0)
                or (
                    quaternions[index][3] == 0
                    and quaternions[index][0] == 0
                    and quaternions[index][1] < 0
                )
                or (
                    quaternions[index][3] == 0
                    and quaternions[index][0] == 0
                    and quaternions[index][1] == 0
                    and quaternions[index][2] < 0
                )
            ):
                quaternions[index][0] *= -1.0
                quaternions[index][1] *= -1.0
                quaternions[index][2] *= -1.0
                quaternions[index][3] *= -1.0

    return quaternions



from torch import Tensor

from ._euler_angle_to_quaternion import (
    euler_angle_to_quaternion,
)
from ._quaternion_magnitude import quaternion_magnitude


def euler_angle_magnitude(
    input: Tensor,
    axes: str,
    degrees: bool | None = False,
) -> Tensor:
    r"""
    Euler angle magnitudes.

    Parameters
    ----------
    input : Tensor, shape (..., 3)
        Euler angles.

    axes : str
        Axes. 1-3 characters belonging to the set {‘X’, ‘Y’, ‘Z’} for intrinsic
        rotations, or {‘x’, ‘y’, ‘z’} for extrinsic rotations. Extrinsic and
        intrinsic rotations cannot be mixed.

    degrees : bool, optional
        If `True`, Euler angles are assumed to be in degrees. Default, `False`.

    Returns
    -------
    euler_angle_magnitudes: Tensor, shape (...)
        Angles in radians. Magnitudes will be in the range :math:`[0, \pi]`.
    """
    return quaternion_magnitude(
        euler_angle_to_quaternion(
            input,
            axes,
            degrees,
        ),
    )



import functools
import operator

import torch
from torch import Tensor

from ._polynomial_vandermonde import polynomial_vandermonde


def polynomial_vandermonde_3d(
    x: Tensor, y: Tensor, z: Tensor, degree: Tensor
) -> Tensor:
    r"""
    Parameters
    ----------
    x : Tensor

    y : Tensor

    z : Tensor

    degree : Tensor

    Returns
    -------
    output : Tensor
    """
    vandermonde_functions = (
        polynomial_vandermonde,
        polynomial_vandermonde,
        polynomial_vandermonde,
    )

    n = len(vandermonde_functions)

    if n != len([x, y, z]):
        raise ValueError

    if n != len(degree):
        raise ValueError

    if n == 0:
        raise ValueError

    matrices = []

    for i in range(n):
        matrix = vandermonde_functions[i]((x, y, z)[i], degree[i])

        matrices = [
            *matrices,
            matrix[(..., *tuple(slice(None) if j == i else None for j in range(n)))],
        ]

    vandermonde = functools.reduce(
        operator.mul,
        matrices,
    )

    return torch.reshape(
        vandermonde,
        [*vandermonde.shape[: -len(degree)], -1],
    )



import torch
from torch import Tensor


def polynomial_companion(input: Tensor) -> Tensor:
    r"""
    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    Returns
    -------
    output : Tensor, shape=(degree, degree)
        Companion matrix.
    """
    input = torch.atleast_1d(input)

    if input.shape[0] < 2:
        raise ValueError

    if input.shape[0] == 2:
        output = torch.tensor([[-input[0] / input[1]]])
    else:
        n = input.shape[0] - 1

        output = torch.reshape(torch.zeros([n, n], dtype=input.dtype), [-1])

        output[n :: n + 1] = 1.0

        output = torch.reshape(output, [n, n])

        output[:, -1] = output[:, -1] + (-input[:-1] / input[-1])

    return output



import beignet
import torch


def test_polynomial_to_chebyshev_polynomial():
    coefficients = [
        torch.tensor([1.0]),
        torch.tensor([0.0, 1]),
        torch.tensor([-1.0, 0, 2]),
        torch.tensor([0.0, -3, 0, 4]),
        torch.tensor([1.0, 0, -8, 0, 8]),
        torch.tensor([0.0, 5, 0, -20, 0, 16]),
        torch.tensor([-1.0, 0, 18, 0, -48, 0, 32]),
        torch.tensor([0.0, -7, 0, 56, 0, -112, 0, 64]),
        torch.tensor([1.0, 0, -32, 0, 160, 0, -256, 0, 128]),
        torch.tensor([0.0, 9, 0, -120, 0, 432, 0, -576, 0, 256]),
    ]

    for index in range(10):
        torch.testing.assert_close(
            beignet.polynomial_to_chebyshev_polynomial(
                coefficients[index],
            ),
            torch.tensor([0.0] * index + [1.0]),
        )



import torch
from torch import Tensor


def rotation_matrix_to_quaternion(
    input: Tensor,
    canonical: bool | None = False,
) -> Tensor:
    r"""
    Convert rotation matrices to rotation quaternions.

    Parameters
    ----------
    input : Tensor, shape=(..., 3, 3)
        Rotation matrices.

    canonical : bool, optional
        Whether to map the redundant double cover of rotation space to a unique
        canonical single cover. If `True`, then the rotation quaternion is
        chosen from :math:`{q, -q}` such that the :math:`w` term is positive.
        If the :math:`w` term is :math:`0`, then the rotation quaternion is
        chosen such that the first non-zero term of the :math:`x`, :math:`y`,
        and :math:`z` terms is positive.

    Returns
    -------
    output : Tensor, shape=(..., 4)
        Rotation quaternion.
    """
    indexes = torch.empty(
        [4],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    output = torch.empty(
        [input.shape[0], 4],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j in range(input.shape[0]):
        indexes[0] = input[j, 0, 0]
        indexes[1] = input[j, 1, 1]
        indexes[2] = input[j, 2, 2]
        indexes[3] = input[j, 0, 0] + input[j, 1, 1] + input[j, 2, 2]

        index, maximum = 0, indexes[0]

        for k in range(1, 4):
            if indexes[k] > maximum:
                index, maximum = k, indexes[k]

        if index == 3:
            output[j, 0] = input[j, 2, 1] - input[j, 1, 2]
            output[j, 1] = input[j, 0, 2] - input[j, 2, 0]
            output[j, 2] = input[j, 1, 0] - input[j, 0, 1]
            output[j, 3] = 1.0 + indexes[3]
        else:
            t = index
            u = (t + 1) % 3
            v = (u + 1) % 3

            output[j, t] = 1.0 - indexes[3] + 2.0 * input[j, t, t]
            output[j, u] = input[j, u, t] + input[j, t, u]
            output[j, v] = input[j, v, t] + input[j, t, v]
            output[j, 3] = input[j, v, u] - input[j, u, v]

        a = output[j, 0] ** 2.0
        b = output[j, 1] ** 2.0
        c = output[j, 2] ** 2.0
        d = output[j, 3] ** 2.0

        output[j] = output[j] / torch.sqrt(a + b + c + d)

    if canonical:
        for j in range(output.shape[0]):
            a = output[j, 0]
            b = output[j, 1]
            c = output[j, 2]
            d = output[j, 3]

            if d == 0 and (a == 0 & (b == 0 & c < 0 | b < 0) | a < 0) | d < 0:
                output[j] = -output[j]

    return output



import importlib
from typing import List


def optional_dependencies(names: List[str], groups: List[str]):
    modules = []

    for name in names:
        try:
            modules = [*modules, importlib.import_module(name)]
        except ImportError:
            modules = [*modules, None]

    missing_names = []

    for name, module in zip(names, modules, strict=False):
        if module is None:
            missing_names = [*missing_names, name]

    if len(missing_names) == 0:
        return

    message = "Missing optional dependencies:\n"

    for missing_name in missing_names:
        message = f"{message}\n    - {missing_name}"

    if len(groups) == 1:
        groups = groups[0]
    else:
        groups = ", ".join(groups)

    message = f"{message}\n\n"

    message = f'{message}Try:\n\n    $ pip install "beignet[{groups}]"'

    raise ImportError(message)



# import beignet
# import torch
#
#
# def test_legendre_polynomial_weight():
#     torch.testing.assert_close(
#         beignet.legendre_polynomial_weight(
#             torch.linspace(-1, 1, 11),
#         ),
#         torch.tensor([1.0]),
#     )



from torch import Tensor


def invert_quaternion(
    input: Tensor,
    canonical: bool = False,
) -> Tensor:
    r"""
    Invert rotation quaternions.

    Parameters
    ----------
    input : Tensor, shape (..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    canonical : bool, optional
        Whether to map the redundant double cover of rotation space to a unique
        canonical single cover. If `True`, then the rotation quaternion is
        chosen from :math:`{q, -q}` such that the :math:`w` term is positive.
        If the :math:`w` term is :math:`0`, then the rotation quaternion is
        chosen such that the first non-zero term of the :math:`x`, :math:`y`,
        and :math:`z` terms is positive.

    Returns
    -------
    inverted_quaternions : Tensor, shape (..., 4)
        Inverted rotation quaternions.
    """
    input[:, :3] = -input[:, :3]

    return input



from pathlib import Path
from typing import Callable, Dict, Tuple, Union

import torch
from pandas import DataFrame
from torch import Tensor

from beignet.transforms import Transform

from ._atom3d_dataset import ATOM3DDataset


class ATOM3DRSRDataset(ATOM3DDataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = False,
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
    ):
        """
        The ATOM3D RNA Structure Ranking (RSR) task predicts the
        three-dimensional structure of an RNA molecule, given its sequence.
        A total of 21 RNAs are included, which consist of the first 21 RNAs
        from the RNA-Puzzles competition (Cruz et al., 2011).

        This problem is prhased as candidate ranking. For each RNA,
        candidate structural models are generated using FARFAR2 (“Silly Boy”
        Watkins et al., 2020) and calculate each candidate’s atoms’ root
        mean squared deviation (RMSD) to the experimentally determined
        structure.

        Each sample is a pair of features and a target, where features is
        the molecule’s atomic coordinates and target is a dictionary of the
        following scores:

        .. list-table:: Target
           :widths: 20 80
           :header-rows: 1

           * - Key
             - Description
           * - score
             -
           * - fa_atr
             -
           * - fa_rep
             -
           * - fa_intra_rep
             -
           * - lk_nonpolar
             -
           * - fa_elec_rna_phos_phos
             -
           * - rna_torsion
             -
           * - suiteness_bonus
             -
           * - rna_sugar_close
             -
           * - fa_stack
             -
           * - stack_elec
             -
           * - geom_sol_fast
             -
           * - hbond_sr_bb_sc
             -
           * - hbond_lr_bb_sc
             -
           * - hbond_sc
             -
           * - ref
             -
           * - free_suite
             -
           * - free_2HOprime
             -
           * - intermol
             -
           * - other_pose
             -
           * - loop_close
             -
           * - linear_chainbreak
             -
           * - rms
             -
           * - rms_stem
             -
           * - time
             -
           * - N_WC
             -
           * - N_NWC
             -
           * - N_BS
             -
           * - N_BP
             -
           * - natWC
             -
           * - natNWC
             -
           * - natBP
             -
           * - f_natWC
             -
           * - f_natNWC
             -
           * - f_natBP
             -

        Parameters
        ----------
        root : Union[str, Path]
            The root directory of the dataset.

        download : bool, optional
            If `True`, download the dataset from the specified source,
            by default `False`.

        transform : Union[Callable, Transform, None], optional
            The transformation function to be applied to the features,
            by default `None`.

        target_transform : Union[Callable, Transform, None], optional
            The transformation function to be applied to the targets,
            by default `None`.
        """
        super().__init__(
            root,
            "raw/candidates/data",
            "https://zenodo.org/record/4961085/files/RSR-raw.tar.gz",
            "RSR",
            checksum="68830ab0ab95cf3d218785a4e4e7669c",
            download=download,
        )

        self._transform_fn = transform

        self._target_transform_fn = target_transform

    def __getitem__(self, index: int) -> Tuple[DataFrame, Dict[str, Tensor]]:
        """
        Parameters
        ----------
        index : int
            The index of the item to retrieve from the dataset.

        Returns
        -------
        Tuple[DataFrame, Dict[str, Tensor]]
            A tuple containing the features and target of the item.
        """
        item = super().__getitem__(index)

        features = DataFrame(**item["atoms"])

        if self._transform_fn is not None:
            features = self._transform_fn(features)

        target = item["scores"]

        for k, v in target.items():
            target[k] = torch.tensor(v)

        if self._target_transform_fn is not None:
            target = self._target_transform_fn(target)

        return features, target



import beignet
import torch


def test_evaluate_physicists_hermite_polynomial_cartesian_3d():
    c1d = torch.tensor([2.5, 1.0, 0.75])
    c3d = torch.einsum(
        "i,j,k->ijk",
        c1d,
        c1d,
        c1d,
    )

    x = torch.rand(3, 5) * 2 - 1
    a, b, x3 = x
    y1, y2, y3 = beignet.evaluate_polynomial(x, torch.tensor([1.0, 2.0, 3.0]))

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial_cartesian_3d(
            a,
            b,
            x3,
            c3d,
        ),
        torch.einsum(
            "i,j,k->ijk",
            y1,
            y2,
            y3,
        ),
    )

    z = torch.ones([2, 3])

    assert (
        beignet.evaluate_physicists_hermite_polynomial_cartesian_3d(z, z, z, c3d).shape
        == (2, 3) * 3
    )



import beignet
import torch


def test_physicists_hermite_polynomial_vandermonde():
    x = torch.arange(3)

    output = beignet.physicists_hermite_polynomial_vandermonde(
        x,
        degree=3,
    )

    assert output.shape == (3, 4)

    for index in range(4):
        torch.testing.assert_close(
            output[..., index],
            beignet.evaluate_physicists_hermite_polynomial(
                x,
                torch.tensor([0.0] * index + [1.0]),
            ),
        )

    output = beignet.physicists_hermite_polynomial_vandermonde(
        torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]),
        degree=3,
    )

    assert output.shape == (3, 2, 4)

    for index in range(4):
        torch.testing.assert_close(
            output[..., index],
            beignet.evaluate_physicists_hermite_polynomial(
                torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]),
                torch.tensor([0.0] * index + [1.0]),
            ),
        )



import torch
from torch import Tensor

from ._add_polynomial import add_polynomial
from ._multiply_polynomial_by_x import multiply_polynomial_by_x
from ._subtract_polynomial import subtract_polynomial


def legendre_polynomial_to_polynomial(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    n = input.shape[0]

    if n < 3:
        return input

    c0 = torch.zeros_like(input)
    c0[0] = input[-2]

    c1 = torch.zeros_like(input)
    c1[0] = input[-1]

    def body(k, c0c1):
        i = n - 1 - k

        c0, c1 = c0c1

        tmp = c0

        c0 = subtract_polynomial(input[i - 2], c1 * (i - 1) / i)

        c1 = add_polynomial(tmp, multiply_polynomial_by_x(c1, "same") * (2 * i - 1) / i)

        return c0, c1

    x = (c0, c1)

    for i in range(0, n - 2):
        x = body(i, x)

    c0, c1 = x

    output = multiply_polynomial_by_x(c1, "same")

    output = add_polynomial(c0, output)

    return output



import math

import beignet
import torch


def test_evaluate_probabilists_hermite_polynomial():
    coefficients = [
        torch.tensor([1]),
        torch.tensor([0, 1]),
        torch.tensor([-1, 0, 1]),
        torch.tensor([0, -3, 0, 1]),
        torch.tensor([3, 0, -6, 0, 1]),
        torch.tensor([0, 15, 0, -10, 0, 1]),
        torch.tensor([-15, 0, 45, 0, -15, 0, 1]),
        torch.tensor([0, -105, 0, 105, 0, -21, 0, 1]),
        torch.tensor([105, 0, -420, 0, 210, 0, -28, 0, 1]),
        torch.tensor([0, 945, 0, -1260, 0, 378, 0, -36, 0, 1]),
    ]

    output = beignet.evaluate_probabilists_hermite_polynomial(
        torch.tensor([]),
        torch.tensor([1.0]),
    )

    assert math.prod(output.shape) == 0

    ys = []

    for coefficient in coefficients:
        ys = [
            *ys,
            beignet.evaluate_polynomial(
                torch.linspace(-1, 1, 50),
                coefficient,
            ),
        ]

    for i in range(10):
        torch.testing.assert_close(
            beignet.evaluate_probabilists_hermite_polynomial(
                torch.linspace(-1, 1, 50),
                torch.tensor([0.0] * i + [1.0]),
            ),
            ys[i],
        )

    for index in range(3):
        shape = (2,) * index

        input = torch.zeros(shape)

        output = beignet.evaluate_probabilists_hermite_polynomial(
            input,
            torch.tensor([1.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_probabilists_hermite_polynomial(
            input,
            torch.tensor([1.0, 0.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_probabilists_hermite_polynomial(
            input,
            torch.tensor([1.0, 0.0, 0.0]),
        )

        assert output.shape == shape



from torch import Tensor

from ._apply_rotation_matrix import apply_rotation_matrix
from ._quaternion_to_rotation_matrix import (
    quaternion_to_rotation_matrix,
)


def apply_quaternion(
    input: Tensor,
    rotation: Tensor,
    inverse: bool | None = False,
) -> Tensor:
    r"""
    Rotates vectors in three-dimensional space using rotation quaternions.

    Note
    ----
    This function interprets the rotation of the original frame to the final
    frame as either a projection, where it maps the components of vectors from
    the final frame to the original frame, or as a physical rotation,
    integrating the vectors into the original frame during the rotation
    process. Consequently, the vector components are maintained in the original
    frame’s perspective both before and after the rotation.

    Parameters
    ----------
    input : Tensor, shape (..., 3)
        Each vector represents a vector in three-dimensional space. The number
        of rotation quaternions and number of vectors must follow standard
        broadcasting rules: either one of them equals unity or they both equal
        each other.

    rotation : Tensor, shape (..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    inverse : bool, optional
        If `True` the inverse of the rotation quaternions are applied to the
        input vectors. Default, `False`.

    Returns
    -------
    output : Tensor, shape (..., 3)
        Rotated vectors.
    """
    return apply_rotation_matrix(
        input,
        quaternion_to_rotation_matrix(
            rotation,
        ),
        inverse,
    )



import beignet
import hypothesis.strategies
import torch
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    rotation = Rotation.random(
        function(
            hypothesis.strategies.integers(
                min_value=16,
                max_value=32,
            ),
        ),
    )

    axes = function(
        hypothesis.strategies.sampled_from(
            [
                "xyz",
                "xzy",
                "yxz",
                "yzx",
                "zxy",
                "zyx",
                "XYZ",
                "XZY",
                "YXZ",
                "YZX",
                "ZXY",
                "ZYX",
            ]
        ),
    )

    degrees = function(hypothesis.strategies.booleans())

    return (
        {
            "input": torch.from_numpy(
                rotation.as_euler(
                    axes,
                    degrees,
                ),
            ),
            "axes": axes,
            "degrees": degrees,
        },
        torch.from_numpy(rotation.as_rotvec(degrees)),
    )


@hypothesis.given(_strategy())
def test_euler_angle_to_rotation_vector(data):
    parameters, expected = data

    torch.testing.assert_close(
        beignet.euler_angle_to_rotation_vector(
            **parameters,
        ),
        expected,
    )



import beignet
import torch


# def test_multiply_chebyshev_polynomial():
#     for j in range(5):
#         for k in range(5):
#             target = torch.zeros(j + k + 1)
#
#             target[abs(j + k)] = target[abs(j + k)] + 0.5
#             target[abs(j - k)] = target[abs(j - k)] + 0.5
#
#             input = torch.tensor([0.0] * j + [1.0])
#             other = torch.tensor([0.0] * k + [1.0])
#
#             torch.testing.assert_close(
#                 beignet.trim_chebyshev_polynomial_coefficients(
#                     beignet.multiply_chebyshev_polynomial(
#                         input,
#                         other,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_chebyshev_polynomial_coefficients(
#                     target,
#                     tol=0.000001,
#                 ),
#             )



# import beignet
# import pytest
# import torch
#
#
# def test_integrate_laguerre_polynomial():
#     with pytest.raises(TypeError):
#         beignet.integrate_laguerre_polynomial(
#             torch.tensor([0]),
#             0.5,
#         )
#
#     with pytest.raises(ValueError):
#         beignet.integrate_laguerre_polynomial(
#             torch.tensor([0]),
#             -1,
#         )
#
#     with pytest.raises(ValueError):
#         beignet.integrate_laguerre_polynomial(
#             torch.tensor([0]),
#             1,
#             torch.tensor([0, 0]),
#         )
#
#     with pytest.raises(ValueError):
#         beignet.integrate_laguerre_polynomial(
#             torch.tensor([0]),
#             lower_bound=[0],
#         )
#
#     with pytest.raises(ValueError):
#         beignet.integrate_laguerre_polynomial(
#             torch.tensor([0]),
#             scale=[0],
#         )
#
#     with pytest.raises(TypeError):
#         beignet.integrate_laguerre_polynomial(
#             torch.tensor([0]),
#             axis=0.5,
#         )
#
#     for i in range(2, 5):
#         torch.testing.assert_close(
#             beignet.trim_laguerre_polynomial_coefficients(
#                 beignet.integrate_laguerre_polynomial(
#                     torch.tensor([0.0]),
#                     order=i,
#                     k=([0.0] * (i - 2) + [1.0]),
#                 ),
#                 tol=0.000001,
#             ),
#             torch.tensor([1.0, -1.0]),
#         )
#
#     for i in range(5):
#         torch.testing.assert_close(
#             beignet.trim_laguerre_polynomial_coefficients(
#                 beignet.laguerre_polynomial_to_polynomial(
#                     beignet.integrate_laguerre_polynomial(
#                         beignet.polynomial_to_laguerre_polynomial(
#                             torch.tensor([0.0] * i + [1.0]),
#                         ),
#                         order=1,
#                         k=[i],
#                     ),
#                 ),
#                 tol=0.000001,
#             ),
#             beignet.trim_laguerre_polynomial_coefficients(
#                 torch.tensor([i] + [0.0] * i + [1.0 / (i + 1.0)]),
#                 tol=0.000001,
#             ),
#         )
#
#     for i in range(5):
#         torch.testing.assert_close(
#             beignet.evaluate_laguerre_polynomial(
#                 torch.tensor([-1.0]),
#                 beignet.integrate_laguerre_polynomial(
#                     beignet.polynomial_to_laguerre_polynomial(
#                         torch.tensor([0.0] * i + [1.0]),
#                     ),
#                     order=1,
#                     k=[i],
#                     lower_bound=-1,
#                 ),
#             ),
#             torch.tensor([i], dtype=torch.get_default_dtype()),
#         )
#
#     for i in range(5):
#         torch.testing.assert_close(
#             beignet.trim_laguerre_polynomial_coefficients(
#                 beignet.laguerre_polynomial_to_polynomial(
#                     beignet.integrate_laguerre_polynomial(
#                         beignet.polynomial_to_laguerre_polynomial(
#                             torch.tensor([0.0] * i + [1.0]),
#                         ),
#                         order=1,
#                         k=[i],
#                         scale=2,
#                     ),
#                 ),
#                 tol=0.000001,
#             ),
#             beignet.trim_laguerre_polynomial_coefficients(
#                 torch.tensor([i] + [0.0] * i + [2.0 / (i + 1)]),
#                 tol=0.000001,
#             ),
#         )
#
#     for i in range(5):
#         for j in range(2, 5):
#             target = torch.tensor([0.0] * i + [1.0])[:]
#
#             for _ in range(j):
#                 target = beignet.integrate_laguerre_polynomial(
#                     target,
#                     order=1,
#                 )
#
#             torch.testing.assert_close(
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     beignet.integrate_laguerre_polynomial(
#                         torch.tensor([0.0] * i + [1.0]),
#                         order=j,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     target,
#                     tol=0.000001,
#                 ),
#             )
#
#     for i in range(5):
#         for j in range(2, 5):
#             target = torch.tensor([0.0] * i + [1.0])[:]
#
#             for k in range(j):
#                 target = beignet.integrate_laguerre_polynomial(
#                     target,
#                     order=1,
#                     k=[k],
#                 )
#
#             torch.testing.assert_close(
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     beignet.integrate_laguerre_polynomial(
#                         torch.tensor([0.0] * i + [1.0]),
#                         order=j,
#                         k=list(range(j)),
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     target,
#                     tol=0.000001,
#                 ),
#             )
#
#     for i in range(5):
#         for j in range(2, 5):
#             target = torch.tensor([0.0] * i + [1.0])[:]
#
#             for k in range(j):
#                 target = beignet.integrate_laguerre_polynomial(
#                     target,
#                     order=1,
#                     k=[k],
#                     lower_bound=-1,
#                 )
#
#             torch.testing.assert_close(
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     beignet.integrate_laguerre_polynomial(
#                         torch.tensor([0.0] * i + [1.0]),
#                         order=j,
#                         k=list(range(j)),
#                         lower_bound=-1,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     target,
#                     tol=0.000001,
#                 ),
#             )
#
#     for i in range(5):
#         for j in range(2, 5):
#             target = torch.tensor([0.0] * i + [1.0])[:]
#
#             for k in range(j):
#                 target = beignet.integrate_laguerre_polynomial(
#                     target,
#                     order=1,
#                     k=[k],
#                     scale=2,
#                 )
#
#             torch.testing.assert_close(
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     beignet.integrate_laguerre_polynomial(
#                         torch.tensor([0.0] * i + [1.0]),
#                         order=j,
#                         k=list(range(j)),
#                         scale=2,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     target,
#                     tol=0.000001,
#                 ),
#             )
#
#     c2d = torch.rand(3, 4)
#
#     torch.testing.assert_close(
#         beignet.integrate_laguerre_polynomial(
#             c2d,
#             axis=0,
#         ),
#         torch.vstack([beignet.integrate_laguerre_polynomial(c) for c in c2d.T]).T,
#     )
#
#     torch.testing.assert_close(
#         beignet.integrate_laguerre_polynomial(
#             c2d,
#             axis=1,
#         ),
#         torch.vstack([beignet.integrate_laguerre_polynomial(c) for c in c2d]),
#     )
#
#     torch.testing.assert_close(
#         beignet.integrate_laguerre_polynomial(
#             c2d,
#             k=3,
#             axis=1,
#         ),
#         torch.vstack([beignet.integrate_laguerre_polynomial(c, k=3) for c in c2d]),
#     )



from torch import Tensor


def lennard_jones_potential(
    input: Tensor,
    sigma: float | Tensor,
    epsilon: float | Tensor,
) -> Tensor:
    r"""
    Lennard-Jones potential.

    Parameters
    ----------
    input : Tensor, shape=(n, m)
        Pairwise distances between particles.

    sigma : float | Tensor, shape=(n, m)
        Distance where the potential energy, :math:`V`, is zero.

    epsilon : float | Tensor, shape=(n, m)
        Depth of the potential well.

    Returns
    -------
    output : Tensor, shape=(n, m)
        Energies.
    """
    a = sigma / input

    b = a**6.0
    c = b**2.0

    return 4.0 * epsilon * (c - b)



import beignet
import torch

torch.set_default_dtype(torch.float64)


def test_multiply_laguerre_polynomial():
    for i in range(5):
        input = torch.linspace(-3, 3, 100)

        a = beignet.evaluate_laguerre_polynomial(
            input,
            torch.tensor([0.0] * i + [1.0]),
        )

        for j in range(5):
            b = beignet.evaluate_laguerre_polynomial(
                input,
                torch.tensor([0.0] * j + [1.0]),
            )

            torch.testing.assert_close(
                beignet.evaluate_laguerre_polynomial(
                    input,
                    beignet.trim_laguerre_polynomial_coefficients(
                        beignet.multiply_laguerre_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                            torch.tensor([0.0] * j + [1.0]),
                        ),
                    ),
                ),
                a * b,
            )


def test_lagline():
    torch.testing.assert_close(
        beignet.linear_laguerre_polynomial(3.0, 4.0),
        torch.tensor([7.0, -4.0]),
    )



import beignet
import pytest
import torch


def test_chebyshev_interpolation():
    def f(x):
        return x * (x - 1) * (x - 2)

    with pytest.raises(ValueError):
        beignet.chebyshev_interpolation(f, -1)

    for i in range(1, 5):
        assert beignet.chebyshev_interpolation(f, i).shape == (i + 1,)

    def powx(x, p):
        return x**p

    x = torch.linspace(-1, 1, 10)

    for i in range(0, 10):
        for j in range(0, i + 1):
            c = beignet.chebyshev_interpolation(
                powx,
                i,
                j,
            )

            torch.testing.assert_close(
                beignet.evaluate_chebyshev_polynomial(x, c),
                powx(x, j),
            )



import functools
from os import PathLike
from pathlib import Path
from typing import Any, Callable

try:
    from mdtraj import Trajectory
except ImportError:
    pass
from torch.utils.data import Dataset


class TrajectoryDataset(Dataset):
    def __init__(
        self,
        func: Callable,
        extension: str,
        root: str | PathLike,
        transform: Callable[["Trajectory"], Any] | None = None,
        stride: int | None = None,
        **kwargs,
    ):
        self.func = functools.partial(func, **kwargs)

        if isinstance(root, str):
            root = Path(root)

        self.root = root.resolve()

        self.transform = transform

        self.stride = stride

        self.paths = [*self.root.glob(f"*.{extension}")]

        super().__init__()

    def __getitem__(self, index: int) -> "Trajectory":
        item = self.func(self.paths[index], stride=self.stride)

        if self.transform:
            item = self.transform(item)

        return item

    def __len__(self) -> int:
        return len(self.paths)



import torch
from torch import Tensor


def laguerre_polynomial_vandermonde(
    x: Tensor,
    degree: Tensor,
) -> Tensor:
    if degree < 0:
        raise ValueError

    x = torch.atleast_1d(x)

    dtype = torch.promote_types(x.dtype, torch.get_default_dtype())

    x = x.to(dtype)

    v = torch.empty([degree + 1, *x.shape], dtype=dtype)

    v[0] = torch.ones_like(x)

    if degree > 0:
        v[1] = 1 - x

        for index in range(2, degree + 1):
            v[index] = (
                v[index - 1] * (2 * index - 1 - x) - v[index - 2] * (index - 1)
            ) / index

    return torch.moveaxis(v, 0, -1)



import torch
import torch.testing
from torch import Tensor


def quaternion_slerp(
    input: Tensor,
    time: Tensor,
    rotation: Tensor,
) -> Tensor:
    r"""
    Interpolate between two or more points on a sphere.

    Unlike linear interpolation, which can result in changes in speed when
    interpolating between orientations or positions on a sphere, spherical
    linear interpolation ensures that the interpolation occurs at a constant
    rate and follows the shortest path on the surface of the sphere.
    The process is useful for rotations and orientation interpolation in
    three-dimensional spaces, smoothly transitioning between different
    orientations.

    Mathematically, spherical linear interpolation interpolates between two
    points on a sphere using a parameter $t$, where $t = 0$ represents the
    start point and $t = n$ represents the end point. For two rotation
    quaternions $q_{1}$ and $q_{2}$ representing the start and end
    orientations:

    $$\text{slerp}(q_{1}, q_{2}; t) = q_{1}\frac{\sin((1 - t)\theta)}{\sin(\theta)} + q_{2}\frac{\sin(t\theta)}{\sin(\theta)}$$

    where $\theta$ is the angle between $q_{1}$ and $q_{2}$, and is computed
    using the dot product of $q_{1}$ and $q_{2}$. This formula ensures that the
    interpolation moves along the shortest path on the four-dimensional sphere
    of rotation quaternions, resulting in a smooth and constant-speed rotation.

    Parameters
    ----------
    input : Tensor, shape (..., N)
        Times.

    time : Tensor, shape (..., N)
        Times of the known rotations. At least 2 times must be specified.

    rotation : Tensor, shape (..., N, 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.
    """  # noqa: E501
    if time.shape[-1] != rotation.shape[-2]:
        raise ValueError

    output = torch.empty(
        [*input.shape, 4],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for index, t in enumerate(input):
        b = torch.min(torch.nonzero(torch.greater_equal(time, t)))

        if b > 0:
            a = b - 1
        else:
            a = b

        if time[b] == t or b == a:
            output[index] = rotation[b]

            continue

        p, q = time[a], time[b]

        r = (t - p) / (q - p)

        t = rotation[a]
        u = rotation[b]

        v = torch.dot(t, u)

        if v < 0.0:
            u = -u
            v = -v

        if v > 0.9995:
            z = (1.0 - r) * t + r * u
        else:
            x = torch.sqrt(1.0 - v**2.0)

            y = torch.atan2(x, v)

            z = t * torch.sin((1.0 - r) * y) / x + u * torch.sin(r * y) / x

        output[index] = z / torch.linalg.norm(z)

    return output



from pathlib import Path
from typing import Callable, TypeVar, Union

from pandas import DataFrame
from torch.utils.data import Dataset

from beignet.transforms import Transform

T = TypeVar("T")


class DataFrameDataset(Dataset):
    _data: DataFrame

    def __init__(
        self,
        root: Union[str, Path],
        *,
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
    ) -> None:
        """
        :param root: Root directory where the dataset subdirectory exists or,
            if :attr:`download` is ``True``, the directory where the dataset
            subdirectory will be created and the dataset downloaded.

        :param transform: A ``Callable`` or ``Transform`` that maps data to
            transformed data (default: ``None``).

        :param target_transform: ``Callable`` or ``Transform`` that maps a
            target to a transformed target (default: ``None``).
        """
        if isinstance(root, str):
            root = Path(root).resolve()

        self._root = root

        self._transform_fn = transform

        self._target_transform_fn = target_transform

    def __getitem__(self, index: int) -> T:
        return self._data.iloc[index]

    def __len__(self) -> int:
        return len(self._data)



import torch
from torch import Tensor


def physicists_hermite_polynomial_vandermonde(
    x: Tensor,
    degree: Tensor,
) -> Tensor:
    if degree < 0:
        raise ValueError

    x = torch.atleast_1d(x)
    dims = (degree + 1,) + x.shape
    dtyp = torch.promote_types(x.dtype, torch.tensor(0.0).dtype)
    x = x.to(dtyp)
    v = torch.empty(dims, dtype=dtyp)
    v[0] = torch.ones_like(x)

    if degree > 0:
        v[1] = x * 2

        for index in range(2, degree + 1):
            v[index] = v[index - 1] * x * 2 - v[index - 2] * (2 * (index - 1))

    return torch.moveaxis(v, 0, -1)



import torch
from torch import Generator, Tensor

from ._quaternion_to_rotation_vector import (
    quaternion_to_rotation_vector,
)
from ._random_quaternion import random_quaternion


def random_rotation_vector(
    size: int,
    degrees: bool = False,
    *,
    generator: Generator | None = None,
    out: Tensor | None = None,
    dtype: torch.dtype | None = None,
    layout: torch.layout | None = torch.strided,
    device: torch.device | None = None,
    requires_grad: bool | None = False,
    pin_memory: bool | None = False,
) -> Tensor:
    r"""
    Generate random rotation vectors.

    Parameters
    ----------
    size : int
        Output size.

    degrees

    generator : torch.Generator, optional
        Psuedo-random number generator. Default, `None`.

    out : Tensor, optional
        Output tensor. Default, `None`.

    dtype : torch.dtype, optional
        Type of the returned tensor. Default, global default.

    layout : torch.layout, optional
        Layout of the returned tensor. Default, `torch.strided`.

    device : torch.device, optional
        Device of the returned tensor. Default, current device for the default
        tensor type.

    requires_grad : bool, optional
        Whether autograd records operations on the returned tensor. Default,
        `False`.

    pin_memory : bool, optional
        If `True`, returned tensor is allocated in pinned memory. Default,
        `False`.

    Returns
    -------
    random_rotation_vectors : Tensor, shape (..., 3)
        Random rotation vectors.
    """
    return quaternion_to_rotation_vector(
        random_quaternion(
            size,
            generator=generator,
            out=out,
            dtype=dtype,
            layout=layout,
            device=device,
            requires_grad=requires_grad,
            pin_memory=pin_memory,
        ),
        degrees,
    )



import beignet
import torch
from torch import Tensor


def test_fit_polynomial():
    def f(x: Tensor) -> Tensor:
        return x * (x - 1) * (x - 2)

    def g(x: Tensor) -> Tensor:
        return x**4 + x**2 + 1

    input = torch.linspace(0, 2, 50)

    other = f(input)

    torch.testing.assert_close(
        beignet.evaluate_polynomial(
            input,
            beignet.fit_polynomial(
                input,
                other,
                degree=3,
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_polynomial(
            input,
            beignet.fit_polynomial(
                input,
                other,
                degree=torch.tensor([0, 1, 2, 3]),
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_polynomial(
            input,
            beignet.fit_polynomial(
                input,
                other,
                degree=4,
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_polynomial(
            input,
            beignet.fit_polynomial(
                input,
                other,
                degree=torch.tensor([0, 1, 2, 3, 4]),
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.fit_polynomial(
            input,
            torch.stack([other, other]).T,
            degree=3,
        ),
        torch.stack(
            [
                beignet.fit_polynomial(
                    input,
                    other,
                    degree=torch.tensor([0, 1, 2, 3]),
                ),
                beignet.fit_polynomial(
                    input,
                    other,
                    degree=torch.tensor([0, 1, 2, 3]),
                ),
            ]
        ).T,
    )

    torch.testing.assert_close(
        beignet.fit_polynomial(
            input,
            torch.stack([other, other]).T,
            degree=torch.tensor([0, 1, 2, 3]),
        ),
        torch.stack(
            [
                beignet.fit_polynomial(
                    input,
                    other,
                    degree=torch.tensor([0, 1, 2, 3]),
                ),
                beignet.fit_polynomial(
                    input,
                    other,
                    degree=torch.tensor([0, 1, 2, 3]),
                ),
            ]
        ).T,
    )

    weight = torch.zeros_like(input)

    weight[1::2] = 1.0

    # torch.testing.assert_close(
    #     beignet.polyfit(
    #         input,
    #         other.at[0::2].set(0),
    #         degree=3,
    #         weight=weight,
    #     ),
    #     beignet.polyfit(
    #         input,
    #         other,
    #         degree=torch.tensor([0, 1, 2, 3]),
    #     ),
    # )
    #
    # torch.testing.assert_close(
    #     beignet.polyfit(
    #         input,
    #         other.at[0::2].set(0),
    #         degree=torch.tensor([0, 1, 2, 3]),
    #         weight=weight,
    #     ),
    #     beignet.polyfit(
    #         input,
    #         other,
    #         degree=torch.tensor([0, 1, 2, 3]),
    #     ),
    # )
    #
    # torch.testing.assert_close(
    #     beignet.polyfit(
    #         input,
    #         torch.tensor([other.at[0::2].set(0), other.at[0::2].set(0)]).T,
    #         degree=3,
    #         weight=weight,
    #     ),
    #     torch.tensor(
    #         [
    #             beignet.polyfit(
    #                 input,
    #                 other,
    #                 degree=torch.tensor([0, 1, 2, 3]),
    #             ),
    #             beignet.polyfit(
    #                 input,
    #                 other,
    #                 degree=torch.tensor([0, 1, 2, 3]),
    #             ),
    #         ]
    #     ).T,
    # )
    #
    # torch.testing.assert_close(
    #     beignet.polyfit(
    #         input,
    #         torch.tensor([other.at[0::2].set(0), other.at[0::2].set(0)]).T,
    #         degree=torch.tensor([0, 1, 2, 3]),
    #         weight=weight,
    #     ),
    #     torch.tensor(
    #         [
    #             beignet.polyfit(
    #                 input,
    #                 other,
    #                 degree=torch.tensor([0, 1, 2, 3]),
    #             ),
    #             beignet.polyfit(
    #                 input,
    #                 other,
    #                 degree=torch.tensor([0, 1, 2, 3]),
    #             ),
    #         ]
    #     ).T,
    # )
    #
    # torch.testing.assert_close(
    #     beignet.polyfit(
    #         torch.tensor([1, 1j, -1, -1j]),
    #         torch.tensor([1, 1j, -1, -1j]),
    #         1,
    #     ),
    #     torch.tensor([0, 1]),
    # )
    #
    # torch.testing.assert_close(
    #     beignet.polyfit(
    #         torch.tensor([1, 1j, -1, -0 - 1j]),
    #         torch.tensor([1, 1j, -1, -0 - 1j]),
    #         (0, 1),
    #     ),
    #     torch.tensor([0, 1]),
    # )
    #
    # input = torch.linspace(-1, 1, 50)
    #
    # other = g(input)
    #
    # torch.testing.assert_close(
    #     beignet.polyval(
    #         input,
    #         beignet.polyfit(
    #             input,
    #             other,
    #             degree=torch.tensor([4]),
    #         ),
    #     ),
    #     other,
    # )
    #
    # torch.testing.assert_close(
    #     beignet.polyval(
    #         input,
    #         beignet.polyfit(
    #             input,
    #             other,
    #             degree=torch.tensor([0, 2, 4]),
    #         ),
    #     ),
    #     other,
    # )
    #
    # torch.testing.assert_close(
    #     beignet.polyfit(
    #         input,
    #         other,
    #         degree=torch.tensor([4]),
    #     ),
    #     beignet.polyfit(
    #         input,
    #         other,
    #         degree=torch.tensor([0, 2, 4]),
    #     ),
    # )



import json
from gzip import GzipFile
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Dict, Union

from torch.utils.data import Dataset

from beignet.transforms import Transform


class LMDBDataset(Dataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        lock: bool = False,
        max_readers: int = 1,
        meminit: bool = True,
        readahead: bool = True,
        readonly: bool = True,
        transform: Union[Callable, Transform, None] = None,
    ):
        super().__init__()

        try:
            import lmdb
        except ImportError as error:
            raise ImportError(
                """
                LMDB datasets require the `lmdb` dependency:

                    $ pip install "beignet[lmdb]"
                """
            ) from error

        self._root = root

        self._transform_fn = transform

        if isinstance(self._root, str):
            self._root = Path(self._root).resolve()

        self._data = lmdb.open(
            str(self._root),
            lock=lock,
            max_readers=max_readers,
            meminit=meminit,
            readahead=readahead,
            readonly=readonly,
        )

        with self._data.begin(write=False) as transaction:
            self._size = int(transaction.get(b"num_examples"))

    def __getitem__(self, index: int) -> Dict[str, Any]:
        if not 0 <= index < self._size:
            raise IndexError(index)

        with self._data.begin(write=False) as transaction:
            with GzipFile(
                fileobj=BytesIO(transaction.get(str(index).encode())),
                mode="rb",
            ) as descriptor:
                item = json.loads(descriptor.read())

        if self._transform_fn:
            item = self._transform_fn(item)

        return item

    def __len__(self) -> int:
        return self._size



from torch import Tensor

from ._evaluate_legendre_polynomial import evaluate_legendre_polynomial


def evaluate_legendre_polynomial_cartesian_2d(
    x: Tensor,
    y: Tensor,
    c: Tensor,
) -> Tensor:
    for arg in [x, y]:
        c = evaluate_legendre_polynomial(arg, c)
    return c



import torch

from ._differentiate_legendre_polynomial import differentiate_legendre_polynomial
from ._evaluate_legendre_polynomial import evaluate_legendre_polynomial
from ._legendre_polynomial_companion import legendre_polynomial_companion


def gauss_legendre_quadrature(degree):
    degree = int(degree)

    if degree <= 0:
        raise ValueError

    c = torch.zeros(degree + 1)
    c[-1] = 1.0
    m = legendre_polynomial_companion(c)
    x = torch.linalg.eigvalsh(m)

    dy = evaluate_legendre_polynomial(x, c)
    df = evaluate_legendre_polynomial(x, differentiate_legendre_polynomial(c))
    x -= dy / df

    fm = evaluate_legendre_polynomial(x, c[1:])

    fm /= torch.abs(fm).max()
    df /= torch.abs(df).max()

    w = 1 / (fm * df)

    a = torch.flip(w, dims=[0])
    b = torch.flip(x, dims=[0])

    w = (w + a) / 2
    x = (x - b) / 2

    w = w * (2.0 / torch.sum(w))

    return x, w



import math

import torch
from torch import Tensor

from ._error_erfi import error_erfi


def dawson_integral_f(input: Tensor, *, out: Tensor | None = None) -> Tensor:
    r"""
    Dawson’s integral.

    Parameters
    ----------
    input : Tensor
        Input tensor.

    out : Tensor, optional
        Output tensor.

    Returns
    -------
    Tensor
    """
    output = math.sqrt(torch.pi) / 2.0 * torch.exp(-(input**2)) * error_erfi(input)

    if out is not None:
        out.copy_(output)

        return out

    return output



import beignet
import pytest
import torch


def test_integrate_probabilists_hermite_polynomial():
    pytest.raises(
        TypeError,
        beignet.integrate_probabilists_hermite_polynomial,
        torch.tensor([0]),
        0.5,
    )
    pytest.raises(
        ValueError,
        beignet.integrate_probabilists_hermite_polynomial,
        torch.tensor([0]),
        -1,
    )
    pytest.raises(
        ValueError,
        beignet.integrate_probabilists_hermite_polynomial,
        torch.tensor([0]),
        1,
        [0, 0],
    )
    pytest.raises(
        ValueError,
        beignet.integrate_probabilists_hermite_polynomial,
        torch.tensor([0]),
        lower_bound=[0],
    )
    pytest.raises(
        ValueError,
        beignet.integrate_probabilists_hermite_polynomial,
        torch.tensor([0]),
        scale=[0],
    )
    pytest.raises(
        TypeError,
        beignet.integrate_probabilists_hermite_polynomial,
        torch.tensor([0]),
        axis=0.5,
    )

    for i in range(2, 5):
        torch.testing.assert_close(
            beignet.trim_probabilists_hermite_polynomial_coefficients(
                beignet.integrate_probabilists_hermite_polynomial(
                    torch.tensor([0.0]),
                    order=i,
                    k=([0.0] * (i - 2) + [1.0]),
                ),
                tol=0.000001,
            ),
            torch.tensor([0.0, 1.0]),
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.trim_probabilists_hermite_polynomial_coefficients(
                beignet.probabilists_hermite_polynomial_to_polynomial(
                    beignet.integrate_probabilists_hermite_polynomial(
                        beignet.polynomial_to_probabilists_hermite_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                        ),
                        order=1,
                        k=[i],
                    ),
                ),
                tol=0.000001,
            ),
            beignet.trim_probabilists_hermite_polynomial_coefficients(
                torch.tensor([i] + [0.0] * i + [1.0 / (i + 1.0)]),
                tol=0.000001,
            ),
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.evaluate_probabilists_hermite_polynomial(
                torch.tensor([-1]),
                beignet.integrate_probabilists_hermite_polynomial(
                    beignet.polynomial_to_probabilists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                    ),
                    order=1,
                    k=[i],
                    lower_bound=-1,
                ),
            ),
            torch.tensor([i], dtype=torch.get_default_dtype()),
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.trim_probabilists_hermite_polynomial_coefficients(
                beignet.probabilists_hermite_polynomial_to_polynomial(
                    beignet.integrate_probabilists_hermite_polynomial(
                        beignet.polynomial_to_probabilists_hermite_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                        ),
                        order=1,
                        k=[i],
                        scale=2,
                    ),
                ),
                tol=0.000001,
            ),
            beignet.trim_probabilists_hermite_polynomial_coefficients(
                torch.tensor([i] + [0.0] * i + [2.0 / (i + 1.0)]),
                tol=0.000001,
            ),
        )

    for i in range(5):
        for j in range(2, 5):
            target = torch.tensor([0.0] * i + [1.0])[:]

            for _ in range(j):
                target = beignet.integrate_probabilists_hermite_polynomial(
                    target,
                    order=1,
                )

            torch.testing.assert_close(
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    beignet.integrate_probabilists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                        order=j,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            target = torch.tensor([0.0] * i + [1.0])[:]
            for k in range(j):
                target = beignet.integrate_probabilists_hermite_polynomial(
                    target,
                    order=1,
                    k=[k],
                )

            torch.testing.assert_close(
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    beignet.integrate_probabilists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                        order=j,
                        k=list(range(j)),
                    ),
                    tol=0.000001,
                ),
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            target = torch.tensor([0.0] * i + [1.0])[:]

            for k in range(j):
                target = beignet.integrate_probabilists_hermite_polynomial(
                    target,
                    order=1,
                    k=[k],
                    lower_bound=-1,
                )

            torch.testing.assert_close(
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    beignet.integrate_probabilists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                        order=j,
                        k=list(range(j)),
                        lower_bound=-1,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            target = torch.tensor([0.0] * i + [1.0])[:]

            for k in range(j):
                target = beignet.integrate_probabilists_hermite_polynomial(
                    target,
                    order=1,
                    k=[k],
                    scale=2,
                )

            torch.testing.assert_close(
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    beignet.integrate_probabilists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                        order=j,
                        k=list(range(j)),
                        scale=2,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_probabilists_hermite_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )

    c2d = torch.rand(3, 4)

    torch.testing.assert_close(
        beignet.integrate_probabilists_hermite_polynomial(
            c2d,
            axis=0,
        ),
        torch.vstack(
            [beignet.integrate_probabilists_hermite_polynomial(c) for c in c2d.T]
        ).T,
    )

    torch.testing.assert_close(
        beignet.integrate_probabilists_hermite_polynomial(
            c2d,
            axis=1,
        ),
        torch.vstack(
            [beignet.integrate_probabilists_hermite_polynomial(c) for c in c2d]
        ),
    )

    torch.testing.assert_close(
        beignet.integrate_probabilists_hermite_polynomial(
            c2d,
            k=3,
            axis=1,
        ),
        torch.vstack(
            [beignet.integrate_probabilists_hermite_polynomial(c, k=3) for c in c2d]
        ),
    )



import beignet
import torch


def test_linear_legendre_polynomial():
    torch.testing.assert_close(
        beignet.linear_legendre_polynomial(3.0, 4.0),
        torch.tensor([3.0, 4.0]),
    )

    torch.testing.assert_close(
        beignet.trim_legendre_polynomial_coefficients(
            beignet.linear_legendre_polynomial(3.0, 0.0),
            tol=0.000001,
        ),
        torch.tensor([3.0]),
    )



import pytest
import torch
from beignet.features import EulerAngle
from torch import Size


class TestEulerAngle:
    @pytest.fixture
    def euler_angles(self) -> EulerAngle:
        return EulerAngle(torch.tensor([[1, 2, 3]]))

    def test___new__(self):
        result = EulerAngle(torch.tensor([[1, 2, 3]]))

        assert isinstance(result, EulerAngle)

        assert result.shape == Size([1, 3])

        with pytest.raises(ValueError):
            EulerAngle(torch.tensor(5))

        with pytest.raises(ValueError):
            EulerAngle(torch.tensor([1, 2]))

    def test___repr__(self, euler_angles: EulerAngle):
        assert isinstance(euler_angles.__repr__(), str)

    def test__wrap(self):
        result = EulerAngle._wrap(torch.tensor([[1, 2, 3]]))

        assert isinstance(result, EulerAngle)

        assert result.shape == Size([1, 3])

    def test_wrap_like(self, euler_angles: EulerAngle):
        result = EulerAngle.wrap_like(
            euler_angles,
            torch.tensor([[1, 2, 3]]),
        )

        assert isinstance(result, EulerAngle)

        assert result.shape == Size([1, 3])



# import beignet
# import torch
#
#
# def test_gauss_laguerre_quadrature():
#     x, w = beignet.gauss_laguerre_quadrature(100)
#
#     v = beignet.lagvander(x, 99)
#     vv = (v.T * w) @ v
#     vd = 1 / torch.sqrt(vv.diagonal())
#     vv = vd[:, None] * vv * vd
#     torch.testing.assert_close(
#         vv,
#         torch.eye(100),
#     )
#
#     target = 1.0
#     torch.testing.assert_close(w.sum(), target)



import math
from typing import Callable

import torch
from torch import Tensor


def chebyshev_interpolation(
    func: Callable[[Tensor, ...], Tensor],
    degree: int,
    *args,
):
    if degree < 0:
        raise ValueError

    order = degree + 1

    if order < 1:
        raise ValueError

    zeros = torch.arange(-order + 1, order + 1, 2)

    zeros = torch.sin(zeros * 0.5 * math.pi / order)

    vandermonde = torch.empty([degree + 1, *zeros.shape], dtype=zeros.dtype)

    vandermonde[0] = torch.ones_like(zeros)

    if degree > 0:
        vandermonde[1] = zeros

        for i in range(2, degree + 1):
            v = vandermonde[i - 1] * zeros * 2.0 - vandermonde[i - 2]

            vandermonde[i] = v

    vandermonde = torch.moveaxis(vandermonde, 0, -1)

    output = vandermonde.T @ func(zeros, *args)

    output[0] = output[0] / order

    output[1:] = output[1:] / (order * 0.5)

    return output



from typing import Literal

import torch
from torch import Tensor
from ._convolve import convolve


def multiply_polynomial(
    input: Tensor,
    other: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    r"""
    Returns the product of two polynomials.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    other : Tensor
        Polynomial coefficients.

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the product.
    """
    input = torch.atleast_1d(input)
    other = torch.atleast_1d(other)

    dtype = torch.promote_types(input.dtype, other.dtype)

    input = input.to(dtype)
    other = other.to(dtype)

    output = convolve(input, other)

    if mode == "same":
        output = output[: max(input.shape[0], other.shape[0])]

    return output



from os import PathLike
from pathlib import Path

from torch.utils.data import Dataset


class SequenceDataset(Dataset):
    def __init__(self, root: str | PathLike, *args, **kwargs):
        if isinstance(root, str):
            root = Path(root)

        self.root = root.resolve()



import heapq
from typing import Any

import torch


def farthest_first_traversal(
    library: list[Any] | torch.Tensor,
    distance_fn: callable,
    ranking_scores: torch.Tensor | None = None,
    n: int = 2,
    descending: bool = False,
) -> torch.Tensor:
    """
    Farthest-first traversal of a library of N elements.
    If `ranking_scores` is provided, the scores are used to pick the starting point
    and to break ties based on edit distance.
    If no scores are provided, ties are broken by the library index.

    Args:
        library: A list with N elements or a tensor with shape (N, *element_shape).
        distance_fn: A callable that takes two arguments and
            returns a distance between them.
        ranking_scores: A tensor with shape (N,) containing the ranking scores
            of the elements in the library.
        n: The number of library elements to return.
        descending: If False and ranking_scores is provided, the elements with
            the lowest scores are selected first in a tie (default: False).

    Returns:
        A tensor with shape (n,) containing the indices of the selected elements.
    """
    if ranking_scores is None:
        remaining_indices = list(range(len(library)))
    else:
        if descending:
            ranking_scores = -ranking_scores
        remaining_indices = list(torch.argsort(ranking_scores))

    selected = [remaining_indices.pop(0)]

    if n == 1:
        return torch.tensor(selected)

    pq = []
    # First pass through library
    for index in remaining_indices:
        # Pushing with heapq, negate dist to simulate max-heap with min-heap
        neg_dist = -distance_fn(library[index], library[selected[0]])

        if ranking_scores is None:
            item = (neg_dist, index, 1)
        else:
            item = (neg_dist, ranking_scores[index], index, 1)

        heapq.heappush(pq, item)

    for _ in range(1, n):
        item = heapq.heappop(pq)
        while True:
            if ranking_scores is None:
                neg_dist, idx, num_checked = item
            else:
                neg_dist, score, idx, num_checked = item

            # Check if the top of the heap has been checked
            # against all currently selected sequences
            if num_checked < len(selected):
                min_dist = min(
                    distance_fn(library[idx], library[selected[i]])
                    for i in range(num_checked, len(selected))
                )
                min_dist = min(min_dist, -neg_dist)

                if ranking_scores is None:
                    item = (-min_dist, idx, len(selected))
                else:
                    item = (-min_dist, score, idx, len(selected))

                item = heapq.heappushpop(pq, item)
            else:
                selected.append(idx)
                break

    return torch.tensor(selected)



from pathlib import Path
from typing import Callable

from beignet.transforms import Transform

from ._tdc_dataset import TDCDataset


class ZINCDataset(TDCDataset):
    def __init__(
        self,
        root: str | Path,
        *,
        download: bool = False,
        transform: Callable | Transform | None = None,
    ):
        r"""
        Parameters
        ----------
        root : str | Path
            Root directory of dataset.

        download: bool
            If `True`, downloads the dataset to the root directory. If dataset
            already exists, it is not redownloaded. Default, `False`.

        transform : Callable | Transform | None
            Transforms the input.
        """
        super().__init__(
            root=root,
            download=download,
            identifier=4170963,
            suffix="tsv",
            checksum="md5:9e4754d72db297d496def3498a926979",
            x_keys=["smiles"],
            y_keys=None,
            transform=transform,
        )



from torch import Tensor

from ._apply_rotation_matrix import apply_rotation_matrix
from ._euler_angle_to_rotation_matrix import euler_angle_to_rotation_matrix


def apply_euler_angle(
    input: Tensor,
    rotation: Tensor,
    axes: str,
    degrees: bool = False,
    inverse: bool = False,
) -> Tensor:
    r"""
    Rotates vectors in three-dimensional space using Euler angles.

    Note
    ----
    This function interprets the rotation of the original frame to the final
    frame as either a projection, where it maps the components of vectors from
    the final frame to the original frame, or as a physical rotation,
    integrating the vectors into the original frame during the rotation
    process. Consequently, the vector components are maintained in the original
    frame’s perspective both before and after the rotation.

    Parameters
    ----------
    input : Tensor
        Vectors in three-dimensional space with the shape $(\ldots \times 3)$.
        Euler angles and vectors must conform to PyTorch broadcasting rules.

    rotation : Tensor
        Euler angles with the shape $(\ldots \times 3)$, specifying the
        rotation in three-dimensional space.

    axes : str
        Specifies the sequence of axes for the rotations, using one to three
        characters from the set ${X, Y, Z}$ for intrinsic rotations, or
        ${x, y, z}$ for extrinsic rotations. Mixing extrinsic and intrinsic
        rotations raises a `ValueError`.

    degrees : bool, optional
        Indicates whether the Euler angles are provided in degrees. If `False`,
        angles are assumed to be in radians. Default, `False`.

    inverse : bool, optional
        If `True`, applies the inverse rotation using the Euler angles to the
        input vectors. Default, `False`.

    Returns
    -------
    output : Tensor
        A tensor of the same shape as `input`, containing the rotated vectors.
    """
    return apply_rotation_matrix(
        input,
        euler_angle_to_rotation_matrix(
            rotation,
            axes,
            degrees,
        ),
        inverse,
    )



import math
import re

import torch
from torch import Tensor


def quaternion_to_euler_angle(
    input: Tensor,
    axes: str,
    degrees: bool = False,
) -> Tensor:
    r"""
    Convert rotation quaternions to Euler angles.

    Parameters
    ----------
    input : Tensor, shape=(..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    axes : str
        Axes. 1-3 characters belonging to the set {‘X’, ‘Y’, ‘Z’} for intrinsic
        rotations, or {‘x’, ‘y’, ‘z’} for extrinsic rotations. Extrinsic and
        intrinsic rotations cannot be mixed.

    degrees : bool, optional
        If `True`, Euler angles are assumed to be in degrees. Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 3)
        Euler angles. The returned Euler angles are in the range:

            * First angle: :math:`(-180, 180]` degrees (inclusive)
            * Second angle:
                * :math:`[-90, 90]` degrees if all axes are different
                  (e.g., :math:`xyz`)
                * :math:`[0, 180]` degrees if first and third axes are the same
                  (e.g., :math:`zxz`)
            * Third angle: :math:`[-180, 180]` degrees (inclusive)
    """
    epsilon = torch.finfo(input.dtype).eps

    output = torch.empty(
        [input.shape[0], 3],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    extrinsic = re.match(r"^[xyz]{1,3}$", axes) is not None

    axes = axes.lower()

    if not extrinsic:
        axes = axes[::-1]

    match axes[0]:
        case "x":
            p = 0
        case "y":
            p = 1
        case "z":
            p = 2
        case _:
            raise ValueError

    match axes[1]:
        case "x":
            q = 0
        case "y":
            q = 1
        case "z":
            q = 2
        case _:
            raise ValueError

    match axes[2]:
        case "x":
            r = 0
        case "y":
            r = 1
        case "z":
            r = 2
        case _:
            raise ValueError

    if p == r:
        r = 3 - p - q

    s = (p - q) * (q - r) * (r - p) // 2

    for j in range(input.shape[0]):
        if p == r:
            t = input[j, 3]
            u = input[j, p]
            v = input[j, q]
            w = input[j, r] * s
        else:
            t = input[j, 3] - input[j, q]
            u = input[j, p] + input[j, r] * s
            v = input[j, q] + input[j, 3]
            w = input[j, r] * s - input[j, p]

        if extrinsic:
            a = 0
            c = 2
        else:
            a = 2
            c = 0

        output[j, 1] = 2.0 * torch.atan2(torch.hypot(v, w), torch.hypot(t, u))

        match output[j, 1]:
            case _ if abs(output[j, 1]) < epsilon:
                output[j, 0] = 2.0 * torch.atan2(u, t)
                output[j, 2] = 0.0
            case _ if abs(output[j, 1] - math.pi) < epsilon:
                if extrinsic:
                    output[j, 0] = 2.0 * -torch.atan2(w, v)
                else:
                    output[j, 0] = 2.0 * +torch.atan2(w, v)

                output[j, 2] = 0.0
            case _:
                output[j, a] = torch.atan2(u, t) - torch.atan2(w, v)
                output[j, c] = torch.atan2(u, t) + torch.atan2(w, v)

        if not p == r:
            output[j, 1] = output[j, 1] - math.pi / 2.0
            output[j, c] = output[j, c] * s

        for k in range(3):
            if output[j, k] <= -math.pi:
                output[j, k] = output[j, k] + math.tau

            if output[j, k] >= +math.pi:
                output[j, k] = output[j, k] - math.tau

    if degrees:
        output = torch.rad2deg(output)

    return output



from pathlib import Path
from typing import Callable

from ..transforms import Transform
from ._uniprot_dataset import UniProtDataset


class UniRef100Dataset(UniProtDataset):
    def __init__(
        self,
        root: str | Path,
        *,
        transform: Callable | Transform | None = None,
        target_transform: Callable | Transform | None = None,
    ):
        r"""
        Parameters
        ----------
        root : str | Path
            Root directory where the dataset subdirectory exists or, if
            `download` is `True`, the directory where the dataset subdirectory
            will be created and the dataset downloaded.

        transform : Callable, optional
            A `Callable` or `Transform` that that maps a sequence to a
            transformed sequence (default: `None`).

        target_transform : Callable, optional
            A `Callable` or `Transform` that maps a target (a cluster
            identifier) to a transformed target (default: `None`).
        """
        super().__init__(
            "http://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref100/uniref100.fasta.gz",
            root,
            "md5:0354240a56f4ca91ff426f8241cfeb7d",
            transform=transform,
            target_transform=target_transform,
        )



import beignet
import torch


def test_laguerre_polynomial_to_polynomial():
    coefficients = [
        torch.tensor([1.0]) / 1,
        torch.tensor([1.0, -1]) / 1,
        torch.tensor([2.0, -4, 1]) / 2,
        torch.tensor([6.0, -18, 9, -1]) / 6,
        torch.tensor([24.0, -96, 72, -16, 1]) / 24,
        torch.tensor([120.0, -600, 600, -200, 25, -1]) / 120,
        torch.tensor([720.0, -4320, 5400, -2400, 450, -36, 1]) / 720,
    ]

    for index in range(7):
        torch.testing.assert_close(
            beignet.laguerre_polynomial_to_polynomial(
                torch.tensor([0.0] * index + [1.0]),
            ),
            coefficients[index],
        )



import beignet
import torch


def test_multiply_laguerre_polynomial_by_x():
    torch.testing.assert_close(
        beignet.trim_laguerre_polynomial_coefficients(
            beignet.multiply_laguerre_polynomial_by_x(
                torch.tensor([0.0]),
            ),
            tol=0.000001,
        ),
        torch.tensor([0.0]),
    )

    torch.testing.assert_close(
        beignet.trim_laguerre_polynomial_coefficients(
            beignet.multiply_laguerre_polynomial_by_x(
                torch.tensor([1.0]),
            ),
            tol=0.000001,
        ),
        torch.tensor([1.0, -1.0]),
    )

    for index in range(1, 5):
        torch.testing.assert_close(
            beignet.trim_laguerre_polynomial_coefficients(
                beignet.multiply_laguerre_polynomial_by_x(
                    torch.tensor([0.0] * index + [1.0]),
                ),
                tol=0.000001,
            ),
            beignet.trim_laguerre_polynomial_coefficients(
                torch.tensor(
                    [0.0] * (index - 1) + [-index, 2.0 * index + 1.0, -(index + 1.0)]
                ),
                tol=0.000001,
            ),
        )



import mdtraj
from beignet.datasets import TrajectoryDataset
from mdtraj import Trajectory


class TestTrajectoryDataset:
    def test___init__(self, data_path):
        dataset = TrajectoryDataset(
            mdtraj.load_pdb,
            "pdb",
            data_path,
        )

        assert dataset.root is not None

    def test___getitem__(self, data_path):
        dataset = TrajectoryDataset(
            mdtraj.load_pdb,
            "pdb",
            data_path,
        )

        assert isinstance(dataset[0], Trajectory)

    def test___len__(self, data_path):
        dataset = TrajectoryDataset(
            mdtraj.load_pdb,
            "pdb",
            data_path,
        )

        assert len(dataset) == 1



import beignet
import hypothesis.extra.numpy
import hypothesis.strategies
import numpy
import torch.testing


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=128,
            max_value=512,
        ),
    )

    input = torch.rand([size])
    other = torch.rand([size])

    return (
        {
            "input": input,
            "other": other,
        },
        torch.reshape(
            torch.from_numpy(numpy.convolve(input, other)),
            [1, -1],
        ),
    )


@hypothesis.given(_strategy())
def test_convolve(data):
    parameters, expected = data

    torch.testing.assert_allclose(
        beignet.convolve(**parameters),
        expected,
    )



import beignet
import torch


def test_laguerre_polynomial_from_roots():
    torch.testing.assert_close(
        beignet.trim_laguerre_polynomial_coefficients(
            beignet.laguerre_polynomial_from_roots(
                torch.tensor([]),
            ),
            tol=0.000001,
        ),
        torch.tensor([1.0]),
    )

    # for i in range(1, 5):
    #     roots = torch.linspace(-math.pi, 0, 2 * i + 1)
    #
    #     roots = roots[1::2]
    #
    #     roots = torch.cos(roots)
    #
    #     output = beignet.lag2poly(
    #         beignet.lagfromroots(
    #             roots,
    #         ),
    #     )
    #
    #     torch.testing.assert_close(
    #         output,
    #         torch.tensor([1.0]),
    #     )
    #
    #     output = beignet.lagval(
    #         roots,
    #         beignet.lagfromroots(
    #             roots,
    #         ),
    #     )
    #
    #     torch.testing.assert_close(
    #         output,
    #         torch.tensor([0.0]),
    #     )



from typing import Callable, Generator

import torch

import beignet
from beignet.transforms import Transform

from ._random_rotation_dataset import RandomRotationDataset


class RandomQuaternionDataset(RandomRotationDataset):
    def __init__(
        self,
        size: int,
        canonical: bool = False,
        *,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        generator: Generator | None = None,
        layout: torch.layout | None = torch.strided,
        pin_memory: bool | None = False,
        requires_grad: bool | None = False,
        transform: Callable | Transform | None = None,
    ):
        r"""
        Parameters
        ----------
        size : int
            Output size.

        canonical : bool, optional
            Whether to map the redundant double cover of rotation space to a
            unique canonical single cover. If `True`, then the rotation
            quaternion is chosen from :math:`{q, -q}` such that the :math:`w`
            term is positive. If the :math:`w` term is :math:`0`, then the
            rotation quaternion is chosen such that the first non-zero term of
            the :math:`x`, :math:`y`, and :math:`z` terms is positive.

        generator : torch.Generator, optional
            Psuedo-random number generator. Default, `None`.

        dtype : torch.dtype, optional
            Type of the returned tensor. Default, global default.

        layout : torch.layout, optional
            Layout of the returned tensor. Default, `torch.strided`.

        device : torch.device, optional
            Device of the returned tensor. Default, current device for the
            default tensor type.

        requires_grad : bool, optional
            Whether autograd records operations on the returned tensor.
            Default, `False`.

        pin_memory : bool, optional
            If `True`, returned tensor is allocated in pinned memory.
            Default, `False`.
        """
        super().__init__(
            beignet.random_quaternion(
                size,
                canonical,
                generator=generator,
                dtype=dtype,
                layout=layout,
                device=device,
                requires_grad=requires_grad,
                pin_memory=pin_memory,
            ),
            transform=transform,
        )



import beignet
import torch


# def test_divide_chebyshev_polynomial():
#     for j in range(5):
#         for k in range(5):
#             input = torch.tensor([0.0] * j + [1.0])
#             other = torch.tensor([0.0] * k + [1.0])
#
#             quotient, remainder = beignet.divide_chebyshev_polynomial(
#                 beignet.add_chebyshev_polynomial(
#                     input,
#                     other,
#                 ),
#                 input,
#             )
#
#             torch.testing.assert_close(
#                 beignet.trim_chebyshev_polynomial_coefficients(
#                     beignet.add_chebyshev_polynomial(
#                         beignet.multiply_chebyshev_polynomial(
#                             quotient,
#                             input,
#                         ),
#                         remainder,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_chebyshev_polynomial_coefficients(
#                     beignet.add_chebyshev_polynomial(
#                         input,
#                         other,
#                     ),
#                     tol=0.000001,
#                 ),
#             )



from typing import Literal

import torch
from torch import Tensor

from ._add_probabilists_hermite_polynomial import add_probabilists_hermite_polynomial
from ._multiply_probabilists_hermite_polynomial_by_x import (
    multiply_probabilists_hermite_polynomial_by_x,
)
from ._subtract_probabilists_hermite_polynomial import (
    subtract_probabilists_hermite_polynomial,
)


def multiply_probabilists_hermite_polynomial(
    input: Tensor,
    other: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    r"""
    Returns the product of two polynomials.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    other : Tensor
        Polynomial coefficients.

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the product.
    """
    input = torch.atleast_1d(input)
    other = torch.atleast_1d(other)

    dtype = torch.promote_types(input.dtype, other.dtype)

    input = input.to(dtype)
    other = other.to(dtype)

    m, n = input.shape[0], other.shape[0]

    if m > n:
        x, y = other, input
    else:
        x, y = input, other

    match x.shape[0]:
        case 1:
            a = add_probabilists_hermite_polynomial(torch.zeros(m + n - 1), x[0] * y)
            b = torch.zeros(m + n - 1)
        case 2:
            a = add_probabilists_hermite_polynomial(torch.zeros(m + n - 1), x[0] * y)
            b = add_probabilists_hermite_polynomial(torch.zeros(m + n - 1), x[1] * y)
        case _:
            size = x.shape[0]

            a = add_probabilists_hermite_polynomial(torch.zeros(m + n - 1), x[-2] * y)
            b = add_probabilists_hermite_polynomial(torch.zeros(m + n - 1), x[-1] * y)

            for i in range(3, x.shape[0] + 1):
                previous = a

                size = size - 1

                a = subtract_probabilists_hermite_polynomial(
                    x[-i] * y, b * (size - 1.0)
                )

                b = add_probabilists_hermite_polynomial(
                    previous, multiply_probabilists_hermite_polynomial_by_x(b, "same")
                )

    output = add_probabilists_hermite_polynomial(
        a, multiply_probabilists_hermite_polynomial_by_x(b, "same")
    )

    if mode == "same":
        output = output[: max(m, n)]

    return output



import operator
import threading
from os import PathLike
from typing import Any, Callable, TypeVar

T = TypeVar("T")


class ThreadSafeFile:
    """
    Share file objects (e.g., raw binary, buffered binary, and text) between
    threads by storing the file object in thread-local storage (TLS).
    """

    def __init__(
        self,
        path: str | PathLike,
        open: Callable[[str | PathLike], T] = operator.methodcaller("open"),
        close: Callable[[T], None] = operator.methodcaller("close"),
    ) -> None:
        self.path = path

        self.open = open

        self.close = close

        self.storage = threading.local()

    def __del__(self) -> None:
        if hasattr(self.storage, "file"):
            self.close(self.storage.file)

            del self.storage.file

    def __getattr__(self, name: str) -> object:
        return getattr(self.file, name)

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()

        if "storage" in state:
            del state["storage"]

        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__ = state

        self.storage = threading.local()

    @property
    def file(self) -> T:
        if not hasattr(self.storage, "file"):
            self.storage.file = self.open(self.path)

        return self.storage.file



import beignet.special
import hypothesis
import hypothesis.strategies
import scipy
import torch


@hypothesis.strategies.composite
def _strategy(function):
    x, y = torch.meshgrid(
        torch.linspace(
            function(
                hypothesis.strategies.floats(
                    min_value=-10,
                    max_value=-10,
                ),
            ),
            function(
                hypothesis.strategies.floats(
                    min_value=10,
                    max_value=10,
                ),
            ),
            steps=128,
            dtype=torch.float64,
        ),
        torch.linspace(
            function(
                hypothesis.strategies.floats(
                    min_value=-10,
                    max_value=-10,
                ),
            ),
            function(
                hypothesis.strategies.floats(
                    min_value=10,
                    max_value=10,
                ),
            ),
            steps=128,
            dtype=torch.float64,
        ),
        indexing="xy",
    )

    input = x + 1.0j * y

    return input, scipy.special.erfc(input)


@hypothesis.given(_strategy())
def test_error_erfc(data):
    input, output = data

    torch.testing.assert_close(beignet.special.error_erfc(input), output)



from torch import Tensor

from ._evaluate_polynomial import evaluate_polynomial


def evaluate_polynomial_cartesian_2d(
    x: Tensor, y: Tensor, coefficients: Tensor
) -> Tensor:
    r"""
    Parameters
    ----------
    x : Tensor

    y : Tensor

    coefficients : Tensor

    Returns
    -------
    output : Tensor
    """
    for input in [x, y]:
        coefficients = evaluate_polynomial(input, coefficients)

    return coefficients



import torch
from torch import Tensor


def differentiate_polynomial(
    input: Tensor,
    order: Tensor | None = None,
    scale: Tensor | None = None,
    dim: int = 0,
) -> Tensor:
    r"""
    Returns the derivative of a polynomial.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    order : Tensor, optional

    scale : Tensor, optional

    dim : int, default=0

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the derivative.
    """
    input = torch.atleast_1d(input)

    if order == 0:
        return input

    input = torch.moveaxis(input, dim, 0)

    if order >= input.shape[0]:
        output = torch.zeros_like(input[:1])
    else:
        d = torch.arange(input.shape[0])

        output = input

        for _ in range(0, order):
            output = (d * output.T).T

            output = torch.roll(output, -1, dims=[0]) * scale

            output[-1] = 0.0

        output = output[:-order]

    output = torch.moveaxis(output, 0, dim)

    return output



import beignet
import torch


def test_multiply_legendre_polynomial_by_x():
    torch.testing.assert_close(
        beignet.trim_legendre_polynomial_coefficients(
            beignet.multiply_legendre_polynomial_by_x(
                torch.tensor([0.0]),
            ),
            tol=0.000001,
        ),
        torch.tensor([0.0]),
    )

    torch.testing.assert_close(
        beignet.trim_legendre_polynomial_coefficients(
            beignet.multiply_legendre_polynomial_by_x(
                torch.tensor([1.0]),
            ),
            tol=0.000001,
        ),
        torch.tensor([0.0, 1.0]),
    )

    for i in range(1, 5):
        torch.testing.assert_close(
            beignet.trim_legendre_polynomial_coefficients(
                beignet.multiply_legendre_polynomial_by_x(
                    torch.tensor([0.0] * i + [1.0]),
                ),
                tol=0.000001,
            ),
            torch.tensor([0] * (i - 1) + [i / (2 * i + 1), 0, (i + 1) / (2 * i + 1)]),
        )



import math

import beignet
import torch


def test_legendre_polynomial_from_roots():
    torch.testing.assert_close(
        beignet.trim_legendre_polynomial_coefficients(
            beignet.legendre_polynomial_from_roots(
                torch.tensor([]),
            ),
            tol=0.000001,
        ),
        torch.tensor([1.0]),
    )

    for index in range(1, 5):
        input = torch.linspace(-math.pi, 0, 2 * index + 1)[1::2]

        output = beignet.legendre_polynomial_from_roots(
            torch.cos(
                input,
            ),
        )

        assert output.shape[-1] == index + 1

        # torch.testing.assert_close(
        #     beignet.leg2poly(
        #         beignet.legfromroots(
        #             torch.cos(
        #                 input,
        #             ),
        #         )
        #     )[-1],
        #     torch.tensor([1.0]),
        # )

        # torch.testing.assert_close(
        #     beignet.legval(
        #         torch.cos(
        #             input,
        #         ),
        #         beignet.legfromroots(
        #             torch.cos(
        #                 input,
        #             ),
        #         ),
        #     ),
        #     torch.tensor([0.0]),
        # )



import beignet
import pytest
import torch


def test_trim_laguerre_polynomial_coefficients():
    with pytest.raises(ValueError):
        beignet.trim_laguerre_polynomial_coefficients(
            torch.tensor([2.0, -1.0, 1.0, 0.0]),
            tol=-1,
        )

    torch.testing.assert_close(
        beignet.trim_laguerre_polynomial_coefficients(
            torch.tensor([2.0, -1.0, 1.0, 0.0]),
        ),
        torch.tensor([2.0, -1.0, 1.0, 0.0])[:-1],
    )

    torch.testing.assert_close(
        beignet.trim_laguerre_polynomial_coefficients(
            torch.tensor([2.0, -1.0, 1.0, 0.0]),
            tol=1,
        ),
        torch.tensor([2.0, -1.0, 1.0, 0.0])[:-3],
    )

    torch.testing.assert_close(
        beignet.trim_laguerre_polynomial_coefficients(
            torch.tensor([2.0, -1.0, 1.0, 0.0]),
            tol=2,
        ),
        torch.tensor([0.0]),
    )



import beignet
import torch


def test_polynomial_to_probabilists_hermite_polynomial():
    coefficients = [
        torch.tensor([1.0]),
        torch.tensor([0.0, 1]),
        torch.tensor([-1.0, 0, 1]),
        torch.tensor([0.0, -3, 0, 1]),
        torch.tensor([3.0, 0, -6, 0, 1]),
        torch.tensor([0.0, 15, 0, -10, 0, 1]),
        torch.tensor([-15.0, 0, 45, 0, -15, 0, 1]),
        torch.tensor([0.0, -105, 0, 105, 0, -21, 0, 1]),
        torch.tensor([105.0, 0, -420, 0, 210, 0, -28, 0, 1]),
        torch.tensor([0.0, 945, 0, -1260, 0, 378, 0, -36, 0, 1]),
    ]

    for index in range(10):
        torch.testing.assert_close(
            beignet.polynomial_to_probabilists_hermite_polynomial(
                coefficients[index],
            ),
            torch.tensor([0.0] * index + [1.0]),
        )



import torch
from torch import Tensor


def euler_angle_to_rotation_matrix(
    input: Tensor,
    axes: str,
    degrees: bool = False,
) -> Tensor:
    r"""
    Convert Euler angles to rotation matrices.

    Parameters
    ----------
    input : Tensor, shape=(..., 3)
        Euler angles.

    axes : str
        Axes. One to three characters belonging to the set :math:`\{X, Y, Z\}`
        for intrinsic rotations, or :math:`\{x, y, z\}` for extrinsic
        rotations. Extrinsic and intrinsic rotations cannot be mixed.

    degrees : bool, optional
        If `True`, Euler angles are assumed to be in degrees. Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 3, 3)
        Rotation matrices.
    """
    if degrees:
        input = torch.deg2rad(input)

    output = torch.empty(
        [input.shape[0], 3, 3],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j, axis in enumerate(axes):
        a = torch.cos(input[..., j])
        b = torch.sin(input[..., j])

        p = torch.full_like(a, 1.0)
        q = torch.full_like(a, 0.0)

        match axis.lower():
            case "x":
                x = [
                    torch.stack([+p, +q, +q], dim=-1),
                    torch.stack([+q, +a, -b], dim=-1),
                    torch.stack([+q, +b, +a], dim=-1),
                ]
            case "y":
                x = [
                    torch.stack([+a, +q, +b], dim=-1),
                    torch.stack([+q, +p, +q], dim=-1),
                    torch.stack([-b, +q, +a], dim=-1),
                ]
            case "z":
                x = [
                    torch.stack([+a, -b, +q], dim=-1),
                    torch.stack([+b, +a, +q], dim=-1),
                    torch.stack([+q, +q, +p], dim=-1),
                ]
            case _:
                raise ValueError

        x = torch.stack(x, dim=-2)

        if j == 0:
            output = x
        else:
            if axes.islower():
                output = x @ output
            else:
                output = output @ x

    return output



import beignet
import torch


def test_chebyshev_polynomial_weight():
    x = torch.linspace(-1, 1, 11)[1:-1]

    torch.testing.assert_close(
        beignet.chebyshev_polynomial_weight(x),
        1.0 / (torch.sqrt(1 + x) * torch.sqrt(1 - x)),
    )
